import asyncio
from unittest import IsolatedAsyncioTestCase

import pytest

from ....admin.server import AdminResponder
from ....messaging.responder import BaseResponder
from ....multitenant.base import BaseMultitenantManager
from ....multitenant.manager import MultitenantManager
from ....tests import mock
from ....utils.testing import create_test_profile
from ...error import WireFormatError
from ...outbound.message import OutboundMessage
from ..message import InboundMessage
from ..receipt import MessageReceipt
from ..session import InboundSession


class TestInboundSession(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

    def test_init(self):
        test_inbound = mock.MagicMock()
        test_session_id = "session-id"
        test_wire_format = mock.MagicMock()
        test_client_info = {"client": "info"}
        test_close = mock.MagicMock()
        test_reply_mode = MessageReceipt.REPLY_MODE_ALL
        test_reply_thread_ids = {"1", "2"}
        test_reply_verkeys = {"3", "4"}
        test_transport_type = "transport-type"
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=test_inbound,
            session_id=test_session_id,
            wire_format=test_wire_format,
            client_info=test_client_info,
            close_handler=test_close,
            reply_mode=test_reply_mode,
            reply_thread_ids=test_reply_thread_ids,
            reply_verkeys=test_reply_verkeys,
            transport_type=test_transport_type,
        )

        assert sess.profile is self.profile
        assert sess.session_id == test_session_id
        assert sess.wire_format is test_wire_format
        assert sess.client_info == test_client_info
        assert sess.reply_mode == test_reply_mode
        assert sess.transport_type == test_transport_type
        assert "1" in sess.reply_thread_ids
        assert "3" in sess.reply_verkeys

        test_msg = mock.MagicMock()
        with mock.patch.object(sess, "process_inbound") as process:
            sess.receive_inbound(test_msg)
            process.assert_called_once_with(test_msg)
            test_inbound.assert_called_once_with(
                sess.profile, test_msg, can_respond=False
            )

        sess.close()
        test_close.assert_called_once()
        assert sess.closed

    def test_setters(self):
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=None,
        )

        sess.reply_mode = MessageReceipt.REPLY_MODE_ALL
        assert sess.reply_mode == MessageReceipt.REPLY_MODE_ALL

        sess.add_reply_thread_ids("1")
        assert "1" in sess.reply_thread_ids
        sess.add_reply_verkeys("2")
        assert "2" in sess.reply_verkeys

        sess.reply_mode = "invalid"
        assert not sess.reply_mode
        assert not sess.reply_thread_ids  # reset by setter method

    async def test_parse_inbound(self):
        test_session_id = "session-id"
        test_transport_type = "transport-type"
        test_wire_format = mock.MagicMock()
        test_wire_format.parse_message = mock.CoroutineMock()
        test_parsed = "parsed-payload"
        test_receipt = mock.MagicMock()
        test_wire_format.parse_message.return_value = (test_parsed, test_receipt)
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=test_session_id,
            wire_format=test_wire_format,
            transport_type=test_transport_type,
        )

        session = self.profile.session()
        setattr(self.profile, "session", mock.MagicMock(return_value=session))

        test_payload = "{}"
        result = await sess.parse_inbound(test_payload)
        test_wire_format.parse_message.assert_awaited_once_with(session, test_payload)
        assert result.payload == test_parsed
        assert result.receipt is test_receipt
        assert result.session_id == test_session_id
        assert result.transport_type == test_transport_type

    async def test_receive(self):
        self.multitenant_mgr = mock.MagicMock(MultitenantManager, autospec=True)
        self.multitenant_mgr.get_wallets_by_message = mock.CoroutineMock(
            return_value=[mock.MagicMock(is_managed=True)]
        )
        self.multitenant_mgr.get_wallet_profile = mock.CoroutineMock(
            return_value=self.profile
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, self.multitenant_mgr
        )
        self.profile.context.update_settings({"multitenant.enabled": True})
        self.base_responder = mock.MagicMock(AdminResponder, autospec=True)
        self.profile.context.injector.bind_instance(BaseResponder, self.base_responder)

        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=None,
        )
        test_msg = mock.MagicMock()

        with (
            mock.patch.object(sess, "parse_inbound", mock.CoroutineMock()) as encode,
            mock.patch.object(sess, "receive_inbound", mock.MagicMock()) as receive,
        ):
            result = await sess.receive(test_msg)
            encode.assert_awaited_once_with(test_msg)
            receive.assert_called_once_with(encode.return_value)
            assert result is encode.return_value

    async def test_receive_no_wallet_found(self):
        self.multitenant_mgr = mock.MagicMock(MultitenantManager, autospec=True)
        self.multitenant_mgr.get_wallets_by_message = mock.CoroutineMock(
            side_effect=ValueError("no such wallet")
        )
        self.multitenant_mgr.get_wallet_profile = mock.CoroutineMock(
            return_value=self.profile
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager, self.multitenant_mgr
        )
        self.profile.context.update_settings({"multitenant.enabled": True})

        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=None,
        )
        test_msg = mock.MagicMock()

        with (
            mock.patch.object(sess, "parse_inbound", mock.CoroutineMock()) as encode,
            mock.patch.object(sess, "receive_inbound", mock.MagicMock()) as receive,
        ):
            result = await sess.receive(test_msg)
            encode.assert_awaited_once_with(test_msg)
            receive.assert_called_once_with(encode.return_value)
            assert result is encode.return_value

    def test_process_inbound(self):
        test_session_id = "session-id"
        test_thread_id = "thread-id"
        test_verkey = "verkey"
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=test_session_id,
            wire_format=None,
        )

        receipt = MessageReceipt(
            direct_response_mode=MessageReceipt.REPLY_MODE_THREAD,
            thread_id=test_thread_id,
            sender_verkey=test_verkey,
        )

        receipt.recipient_did = "dummy"
        assert receipt.recipient_did == "dummy"
        receipt.recipient_did_public = True
        assert receipt.recipient_did_public
        receipt.recipient_did = None
        receipt.recipient_did_public = None
        assert receipt.recipient_did is None
        assert receipt.recipient_did_public is None
        receipt.sender_did = "dummy"
        assert receipt.sender_did == "dummy"
        receipt.sender_did = None
        assert receipt.sender_did is None
        assert "direct_response_mode" in str(receipt)

        message = InboundMessage(payload=None, receipt=receipt)
        sess.process_inbound(message)
        assert sess.reply_mode == receipt.direct_response_mode
        assert test_verkey in sess.reply_verkeys
        assert test_thread_id in sess.reply_thread_ids

        assert receipt.in_time is None
        receipt.connection_id = "dummy"
        assert receipt.connection_id == "dummy"

    def test_select_outbound(self):
        test_session_id = "session-id"
        test_thread_id = "thread-id"
        test_verkey = "verkey"
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=test_session_id,
            wire_format=None,
        )

        sess.reply_mode = MessageReceipt.REPLY_MODE_ALL
        test_msg = OutboundMessage(payload=None)
        assert not sess.select_outbound(test_msg)  # no key
        test_msg.reply_session_id = test_session_id
        assert not sess.select_outbound(test_msg)  # no difference
        sess.can_respond = True
        assert not sess.select_outbound(test_msg)  # no difference
        test_msg.reply_to_verkey = test_verkey
        sess.add_reply_verkeys(test_verkey)
        assert sess.select_outbound(test_msg)

        sess.reply_mode = MessageReceipt.REPLY_MODE_THREAD
        sess.reply_verkeys = None
        sess.reply_thread_ids = None
        test_msg = OutboundMessage(payload=None)
        assert not sess.select_outbound(test_msg)
        sess.add_reply_thread_ids(test_thread_id)
        test_msg.reply_thread_id = test_thread_id
        assert not sess.select_outbound(test_msg)
        sess.add_reply_verkeys(test_verkey)
        test_msg.reply_to_verkey = test_verkey
        assert sess.select_outbound(test_msg)

        sess.close()
        assert not sess.select_outbound(test_msg)

    async def test_wait_response(self):
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=None,
        )
        test_msg = OutboundMessage(payload=None)
        sess.set_response(test_msg)
        assert sess.response_event.is_set()
        assert sess.response_buffered

        with mock.patch.object(sess, "encode_outbound", mock.CoroutineMock()) as encode:
            result = await asyncio.wait_for(sess.wait_response(), 0.1)
            encode.assert_awaited_once_with(test_msg)
            assert result is encode.return_value

        sess.clear_response()
        assert not sess.response_buffer

        sess.close()
        assert await asyncio.wait_for(sess.wait_response(), 0.1) is None

    async def test_wait_response_x(self):
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=None,
        )
        test_msg = OutboundMessage(payload=None)
        sess.set_response(test_msg)
        assert sess.response_event.is_set()
        assert sess.response_buffered

        with mock.patch.object(sess, "encode_outbound", mock.CoroutineMock()) as encode:
            encode.side_effect = WireFormatError()
            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(sess.wait_response(), 0.1)

        assert not sess.response_buffer

        sess.close()
        assert await asyncio.wait_for(sess.wait_response(), 0.1) is None

    async def test_encode_response(self):
        test_wire_format = mock.MagicMock()
        test_wire_format.encode_message = mock.CoroutineMock()
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=test_wire_format,
        )
        test_msg = OutboundMessage(payload=None)
        test_from_verkey = "from-verkey"
        test_to_verkey = "to-verkey"

        session = self.profile.session()
        setattr(self.profile, "session", mock.MagicMock(return_value=session))

        with self.assertRaises(WireFormatError):
            await sess.encode_outbound(test_msg)
        test_msg.payload = "{}"
        with self.assertRaises(WireFormatError):
            await sess.encode_outbound(test_msg)
        test_msg.reply_from_verkey = test_from_verkey
        test_msg.reply_to_verkey = test_to_verkey
        result = await sess.encode_outbound(test_msg)
        assert result is test_wire_format.encode_message.return_value

        test_wire_format.encode_message.assert_awaited_once_with(
            session,
            test_msg.payload,
            [test_to_verkey],
            None,
            test_from_verkey,
        )

    async def test_accept_response(self):
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=None,
        )
        test_msg = OutboundMessage(payload=None)

        with mock.patch.object(sess, "select_outbound") as selector:
            selector.return_value = False

            accepted = sess.accept_response(test_msg)
            assert not accepted and not accepted.retry

            sess.set_response(OutboundMessage(payload=None))
            selector.return_value = True
            accepted = sess.accept_response(test_msg)
            assert not accepted and accepted.retry

            sess.clear_response()
            accepted = sess.accept_response(test_msg)
            assert accepted

    async def test_context_mgr(self):
        sess = InboundSession(
            profile=self.profile,
            inbound_handler=None,
            session_id=None,
            wire_format=None,
        )
        assert not sess.closed
        async with sess:
            pass
        assert sess.closed



from unittest import TestCase, mock

import pytest

from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import PRES_20_PROBLEM_REPORT, PROTOCOL_PACKAGE
from ..pres_problem_report import ProblemReportReason, V20PresProblemReport


class TestV20PresProblemReport(TestCase):
    """Problem report tests."""

    def test_init_type(self):
        """Test initializer."""

        prob = V20PresProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )
        assert prob._type == DIDCommPrefix.qualify_current(PRES_20_PROBLEM_REPORT)

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.pres_problem_report.V20PresProblemReportSchema.load"
    )
    def test_deserialize(self, mock_load):
        """Test deserialization."""

        obj = V20PresProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )

        prob = V20PresProblemReport.deserialize(obj)
        mock_load.assert_called_once_with(obj)

        assert prob is mock_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.pres_problem_report.V20PresProblemReportSchema.dump"
    )
    def test_serialize(self, mock_dump):
        """Test serialization."""

        obj = V20PresProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )

        ser = obj.serialize()
        mock_dump.assert_called_once_with(obj)

        assert ser is mock_dump.return_value

    def test_make_model(self):
        """Test making model."""

        prob = V20PresProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ABANDONED.value,
            }
        )
        data = prob.serialize()
        model_instance = V20PresProblemReport.deserialize(data)
        assert isinstance(model_instance, V20PresProblemReport)

        prob = V20PresProblemReport()
        data = prob.serialize()
        with pytest.raises(BaseModelError):
            V20PresProblemReport.deserialize(data)



from unittest import TestCase, mock

import pytest

from ......messaging.models.base import BaseModelError
from .....didcomm_prefix import DIDCommPrefix
from ...message_types import CREDENTIAL_PROBLEM_REPORT, PROTOCOL_PACKAGE
from .. import credential_problem_report as test_module
from ..credential_problem_report import (
    CredentialProblemReport,
    CredentialProblemReportSchema,
    ProblemReportReason,
    ValidationError,
)


class TestCredentialProblemReport(TestCase):
    """Problem report tests."""

    def test_init_type(self):
        """Test initializer."""

        prob = CredentialProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )
        assert prob._type == DIDCommPrefix.qualify_current(CREDENTIAL_PROBLEM_REPORT)

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.credential_problem_report."
        "CredentialProblemReportSchema.load"
    )
    def test_deserialize(self, mock_load):
        """Test deserialization."""

        obj = CredentialProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )

        prob = CredentialProblemReport.deserialize(obj)
        mock_load.assert_called_once_with(obj)

        assert prob is mock_load.return_value

    @mock.patch(
        f"{PROTOCOL_PACKAGE}.messages.credential_problem_report."
        "CredentialProblemReportSchema.dump"
    )
    def test_serialize(self, mock_dump):
        """Test serialization."""

        obj = CredentialProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )

        ser = obj.serialize()
        mock_dump.assert_called_once_with(obj)

        assert ser is mock_dump.return_value

    def test_make_model(self):
        """Test making model."""

        prob = CredentialProblemReport(
            description={
                "en": "oh no",
                "code": ProblemReportReason.ISSUANCE_ABANDONED.value,
            }
        )
        data = prob.serialize()
        model_instance = CredentialProblemReport.deserialize(data)
        assert isinstance(model_instance, CredentialProblemReport)

        prob = CredentialProblemReport()
        data = prob.serialize()
        with pytest.raises(BaseModelError):
            CredentialProblemReport.deserialize(data)

    def test_validate_x(self):
        """Exercise validation requirements."""
        schema = CredentialProblemReportSchema()
        with pytest.raises(ValidationError):
            schema.validate_fields({})

    def test_validate_and_logger(self):
        """Capture ValidationError and Logs."""
        data = CredentialProblemReport(
            description={
                "en": "oh no",
                "code": "invalid_code",
            },
        ).serialize()
        with mock.patch.object(test_module, "LOGGER", autospec=True) as mock_logger:
            CredentialProblemReportSchema().validate_fields(data)
        assert mock_logger.warning.call_count == 1



"""Test KeyDIDResolver."""

import pytest
import pytest_asyncio

from ....core.profile import Profile
from ....messaging.valid import DIDKey
from ....utils.testing import create_test_profile
from ...base import DIDNotFound
from ..key import KeyDIDResolver

# pylint: disable=W0621
TEST_DID0 = "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th"
TEST_DID_INVALID = "did:key:z1MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th"


@pytest.fixture
def resolver():
    """Resolver fixture."""
    yield KeyDIDResolver()


@pytest_asyncio.fixture
async def profile():
    """Profile fixture."""
    profile = await create_test_profile()
    yield profile


@pytest.mark.asyncio
async def test_supported_did_regex(profile, resolver: KeyDIDResolver):
    """Test the supported_did_regex."""
    assert resolver.supported_did_regex == DIDKey.PATTERN
    assert await resolver.supports(
        profile,
        TEST_DID0,
    )


@pytest.mark.asyncio
async def test_resolve(resolver: KeyDIDResolver, profile: Profile):
    """Test resolve method."""
    assert await resolver.resolve(profile, TEST_DID0)


@pytest.mark.asyncio
async def test_resolve_x_did_not_found(resolver: KeyDIDResolver, profile: Profile):
    """Test resolve method when no did is found."""
    with pytest.raises(DIDNotFound):
        await resolver.resolve(profile, TEST_DID_INVALID)



"""Handle registration and publication of supported goal codes."""

from typing import Sequence

from ..utils.classloader import ClassLoader


class GoalCodeRegistry:
    """Goal code registry."""

    def __init__(self):
        """Initialize a `GoalCodeRegistry` instance."""
        self.goal_codes = []

    def register_controllers(self, *controller_sets):
        """Add new controllers.

        Args:
            controller_sets: Mappings of controller to coroutines

        """
        for controlset in controller_sets:
            for key, ctl_cls in controlset.items():
                ctl_cls = ClassLoader.load_class(ctl_cls)
                ctl_inst = ctl_cls(protocol=key)
                goal_codes_to_add = ctl_inst.determine_goal_codes()
                for goal_code in goal_codes_to_add:
                    if goal_code not in self.goal_codes:
                        self.goal_codes.append(goal_code)

    def goal_codes_matching_query(self, query: str) -> Sequence[str]:
        """Return a list of goal codes matching a query string."""
        all_types = self.goal_codes
        result = None

        if query == "*" or query is None:
            result = all_types
        elif query:
            if query.endswith("*"):
                match = query[:-1]
                result = tuple(k for k in all_types if k.startswith(match))
            elif query in all_types:
                result = (query,)
        return result or ()



import pytest
import pytest_asyncio

from ....storage.error import StorageDuplicateError, StorageNotFoundError
from ....utils.testing import create_test_profile
from ..base import VCHolder
from ..vc_record import VCRecord

VC_CONTEXT = "https://www.w3.org/2018/credentials/v1"
VC_TYPE = "https://www.w3.org/2018/credentials#VerifiableCredential"
VC_SUBJECT_ID = "did:example:ebfeb1f712ebc6f1c276e12ec21"
VC_PROOF_TYPE = "Ed25519Signature2018"
VC_ISSUER_ID = "https://example.edu/issuers/14"
VC_SCHEMA_ID = "https://example.org/examples/degree.json"
VC_GIVEN_ID = "http://example.edu/credentials/3732"


@pytest_asyncio.fixture
async def holder():
    profile = await create_test_profile(settings={"wallet.type": "askar"})
    yield profile.inject(VCHolder)


@pytest.fixture
def record():
    yield VCRecord(
        contexts=[
            VC_CONTEXT,
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        expanded_types=[
            VC_TYPE,
            "https://example.org/examples#UniversityDegreeCredential",
        ],
        schema_ids=[VC_SCHEMA_ID],
        issuer_id=VC_ISSUER_ID,
        subject_ids=[VC_SUBJECT_ID],
        proof_types=[VC_PROOF_TYPE],
        given_id=VC_GIVEN_ID,
        cred_tags={"tag": "value"},
        cred_value={
            "@context": [
                VC_CONTEXT,
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "id": VC_GIVEN_ID,
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "issuer": VC_ISSUER_ID,
            "identifier": "83627467",
            "name": "University Degree",
            "issuanceDate": "2010-01-01T19:53:24Z",
            "credentialSubject": {
                "id": VC_SUBJECT_ID,
                "givenName": "Cai",
                "familyName": "Leblanc",
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2021-05-07T08:50:17.626625",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..rubQvgig7cN-F6cYn_AJF1BCSaMpkoR517Ot_4pqwdJnQ-JwKXq6d6cNos5JR73E9WkwYISXapY0fYTIG9-fBA",
            },
        },
    )


@pytest.mark.asyncio
async def test_tag_query(holder: VCHolder, record: VCRecord):
    test_uri_list = [
        "https://www.w3.org/2018/credentials#VerifiableCredential",
        "https://example.org/examples#UniversityDegreeCredential",
    ]
    test_query = holder.build_type_or_schema_query(test_uri_list)
    assert test_query == {
        "$and": [
            {
                "$or": [
                    {"type": "https://www.w3.org/2018/credentials#VerifiableCredential"},
                    {
                        "schema": "https://www.w3.org/2018/credentials#VerifiableCredential"
                    },
                ]
            },
            {
                "$or": [
                    {"type": "https://example.org/examples#UniversityDegreeCredential"},
                    {"schema": "https://example.org/examples#UniversityDegreeCredential"},
                ]
            },
        ]
    }
    await holder.store_credential(record)

    search = holder.search_credentials(pd_uri_list=test_uri_list)
    rows = await search.fetch()
    assert rows == [record]

    @pytest.mark.asyncio
    async def test_handle_parser_error(holder: VCHolder):
        record = VCRecord(
            contexts=[
                VC_CONTEXT,
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            expanded_types=[
                VC_TYPE,
                "https://example.org/examples#UniversityDegreeCredential",
            ],
            schema_ids=[VC_SCHEMA_ID],
            issuer_id=VC_ISSUER_ID,
            subject_ids=[VC_SUBJECT_ID],
            proof_types=[VC_PROOF_TYPE],
            given_id=VC_GIVEN_ID,
            cred_tags={"tag": "value"},
            cred_value={
                "@context": [
                    VC_CONTEXT,
                    "https://www.w3.org/2018/credentials/examples/v1",
                ],
                "id": VC_GIVEN_ID,
                "type": ["VerifiableCredential", "UniversityDegreeCredential"],
                "issuer": VC_ISSUER_ID,
                "identifier": "83627467",
                "name": "University Degree",
                "issuanceDate": "20180-10-29T21:02:19.201+0000",
                "credentialSubject": {
                    "id": VC_SUBJECT_ID,
                    "givenName": "Cai",
                    "familyName": "Leblanc",
                },
                "proof": {
                    "type": "Ed25519Signature2018",
                    "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "created": "2021-05-07T08:50:17.626625",
                    "proofPurpose": "assertionMethod",
                    "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..rubQvgig7cN-F6cYn_AJF1BCSaMpkoR517Ot_4pqwdJnQ-JwKXq6d6cNos5JR73E9WkwYISXapY0fYTIG9-fBA",
                },
            },
        )
        await holder.store_credential(record)
        search = holder.search_credentials()
        rows = await search.fetch()
        assert rows == [record]


@pytest.mark.asyncio
async def test_sorting_vcrecord(holder: VCHolder):
    record_a = VCRecord(
        contexts=[
            VC_CONTEXT,
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        expanded_types=[
            VC_TYPE,
            "https://example.org/examples#UniversityDegreeCredential",
        ],
        schema_ids=[VC_SCHEMA_ID],
        issuer_id=VC_ISSUER_ID,
        subject_ids=[VC_SUBJECT_ID],
        proof_types=[VC_PROOF_TYPE],
        given_id=VC_GIVEN_ID,
        cred_tags={"tag": "value"},
        cred_value={
            "@context": [
                VC_CONTEXT,
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "id": VC_GIVEN_ID,
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "issuer": VC_ISSUER_ID,
            "identifier": "83627467",
            "name": "University Degree",
            "issuanceDate": "2010-01-01T19:53:24Z",
            "credentialSubject": {
                "id": VC_SUBJECT_ID,
                "givenName": "Cai",
                "familyName": "Leblanc",
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2021-05-07T08:50:17.626625",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..rubQvgig7cN-F6cYn_AJF1BCSaMpkoR517Ot_4pqwdJnQ-JwKXq6d6cNos5JR73E9WkwYISXapY0fYTIG9-fBA",
            },
        },
    )
    await holder.store_credential(record_a)
    record_b = VCRecord(
        contexts=[
            VC_CONTEXT,
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        expanded_types=[
            VC_TYPE,
            "https://example.org/examples#UniversityDegreeCredential",
        ],
        schema_ids=[VC_SCHEMA_ID],
        issuer_id=VC_ISSUER_ID,
        subject_ids=[VC_SUBJECT_ID],
        proof_types=[VC_PROOF_TYPE],
        given_id=VC_GIVEN_ID,
        cred_tags={"tag": "value"},
        cred_value={
            "@context": [
                VC_CONTEXT,
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "id": VC_GIVEN_ID,
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "issuer": VC_ISSUER_ID,
            "identifier": "83627467",
            "name": "University Degree",
            "issuanceDate": "2012-01-01T19:53:24Z",
            "credentialSubject": {
                "id": VC_SUBJECT_ID,
                "givenName": "Cai",
                "familyName": "Leblanc",
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2021-05-07T08:50:17.626625",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..rubQvgig7cN-F6cYn_AJF1BCSaMpkoR517Ot_4pqwdJnQ-JwKXq6d6cNos5JR73E9WkwYISXapY0fYTIG9-fBA",
            },
        },
    )
    await holder.store_credential(record_b)
    record_c = VCRecord(
        contexts=[
            VC_CONTEXT,
            "https://www.w3.org/2018/credentials/examples/v1",
        ],
        expanded_types=[
            VC_TYPE,
            "https://example.org/examples#UniversityDegreeCredential",
        ],
        schema_ids=[VC_SCHEMA_ID],
        issuer_id=VC_ISSUER_ID,
        subject_ids=[VC_SUBJECT_ID],
        proof_types=[VC_PROOF_TYPE],
        given_id=VC_GIVEN_ID,
        cred_tags={"tag": "value"},
        cred_value={
            "@context": [
                VC_CONTEXT,
                "https://www.w3.org/2018/credentials/examples/v1",
            ],
            "id": VC_GIVEN_ID,
            "type": ["VerifiableCredential", "UniversityDegreeCredential"],
            "issuer": VC_ISSUER_ID,
            "identifier": "83627467",
            "name": "University Degree",
            "issuanceDate": "2009-01-01T19:53:24Z",
            "credentialSubject": {
                "id": VC_SUBJECT_ID,
                "givenName": "Cai",
                "familyName": "Leblanc",
            },
            "proof": {
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "created": "2021-05-07T08:50:17.626625",
                "proofPurpose": "assertionMethod",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..rubQvgig7cN-F6cYn_AJF1BCSaMpkoR517Ot_4pqwdJnQ-JwKXq6d6cNos5JR73E9WkwYISXapY0fYTIG9-fBA",
            },
        },
    )
    await holder.store_credential(record_c)
    # expected = [record_b, record_a, record_c]

    search = holder.search_credentials()
    rows = await search.fetch()
    assert rows  # == expected


@pytest.mark.asyncio
async def test_tag_query_valid_and_operator(holder: VCHolder, record: VCRecord):
    test_uri_list = [
        "https://www.w3.org/2018/credentials#VerifiableCredential",
        "https://example.org/examples#UniversityDegreeCredential2",
    ]
    await holder.store_credential(record)

    search = holder.search_credentials(pd_uri_list=test_uri_list)
    rows = await search.fetch()
    assert rows == []


@pytest.mark.asyncio
async def test_store_retrieve(holder: VCHolder, record: VCRecord):
    await holder.store_credential(record)
    result = await holder.retrieve_credential_by_id(record.record_id)
    assert result == record

    result = await holder.retrieve_credential_by_given_id(record.given_id)
    assert result == record

    with pytest.raises(StorageDuplicateError):
        await holder.store_credential(record)

    with pytest.raises(StorageNotFoundError):
        await holder.retrieve_credential_by_id("missing")

    with pytest.raises(StorageNotFoundError):
        await holder.retrieve_credential_by_given_id("missing")


@pytest.mark.asyncio
async def test_delete(holder: VCHolder, record: VCRecord):
    await holder.store_credential(record)
    await holder.delete_credential(record)
    with pytest.raises(StorageNotFoundError):
        await holder.retrieve_credential_by_id(record.record_id)


@pytest.mark.asyncio
async def test_search(holder: VCHolder, record: VCRecord):
    await holder.store_credential(record)

    search = holder.search_credentials()
    rows = await search.fetch()
    assert rows == [record]
    await search.close()

    # test async iter and repr
    search = holder.search_credentials()
    assert search.__class__.__name__ in str(search)
    rows = []
    async for row in search:
        rows.append(row)
    assert rows == [record]
    await search.close()

    search = holder.search_credentials(
        contexts=[VC_CONTEXT],
        types=[VC_TYPE],
        schema_ids=[VC_SCHEMA_ID],
        subject_ids=[VC_SUBJECT_ID],
        proof_types=[VC_PROOF_TYPE],
        issuer_id=VC_ISSUER_ID,
        given_id=VC_GIVEN_ID,
        tag_query={"tag": "value"},
    )
    rows = await search.fetch()
    assert rows == [record]

    rows = await holder.search_credentials(contexts=["other-context"]).fetch()
    assert not rows

    rows = await holder.search_credentials(types=["other-type"]).fetch()
    assert not rows

    rows = await holder.search_credentials(schema_ids=["other schema"]).fetch()
    assert not rows

    rows = await holder.search_credentials(subject_ids=["other subject"]).fetch()
    assert not rows

    rows = await holder.search_credentials(proof_types=["other proof type"]).fetch()
    assert not rows

    rows = await holder.search_credentials(issuer_id="other issuer").fetch()
    assert not rows

    rows = await holder.search_credentials(given_id="other given id").fetch()
    assert not rows

    await search.close()



"""V2.0 present-proof indy presentation-exchange format handler."""

import json
import logging
from typing import Mapping, Optional, Tuple

from marshmallow import RAISE

from ......core.profile import Profile
from ......indy.holder import IndyHolder
from ......indy.models.predicate import Predicate
from ......indy.models.proof import IndyProofSchema
from ......indy.models.proof_request import IndyProofRequestSchema
from ......indy.models.xform import indy_proof_req_preview2indy_requested_creds
from ......indy.util import generate_pr_nonce
from ......indy.verifier import IndyVerifier
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......messaging.util import canon
from ....indy.pres_exch_handler import IndyPresExchHandler
from ...message_types import ATTACHMENT_FORMAT, PRES_20, PRES_20_PROPOSAL, PRES_20_REQUEST
from ...messages.pres import V20Pres
from ...messages.pres_format import V20PresFormat
from ...models.pres_exchange import V20PresExRecord
from ..anoncreds.handler import AnonCredsPresExchangeHandler
from ..handler import V20PresFormatHandler, V20PresFormatHandlerError

LOGGER = logging.getLogger(__name__)


class IndyPresExchangeHandler(V20PresFormatHandler):
    """Indy presentation format handler."""

    format = V20PresFormat.Format.INDY

    def __init__(self, profile: Profile):
        """Shim initialization to check for new AnonCreds library."""
        super().__init__(profile)

    @classmethod
    def validate_fields(cls, message_type: str, attachment_data: Mapping):
        """Validate attachment data for a specific message type.

        Uses marshmallow schemas to validate if format specific attachment data
        is valid for the specified message type. Only does structural and type
        checks, does not validate if .e.g. the issuer value is valid.


        Args:
            message_type (str): The message type to validate the attachment data for.
                Should be one of the message types as defined in message_types.py
            attachment_data (Mapping): [description]
                The attachment data to validate

        Raises:
            Exception: When the data is not valid.

        """
        mapping = {
            PRES_20_REQUEST: IndyProofRequestSchema,
            PRES_20_PROPOSAL: IndyProofRequestSchema,
            PRES_20: IndyProofSchema,
        }

        # Get schema class
        Schema = mapping[message_type]

        # Validate, throw if not valid
        Schema(unknown=RAISE).load(attachment_data)

    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier

        Returns:
            str: Issue credential attachment format identifier

        """

        return ATTACHMENT_FORMAT[message_type][IndyPresExchangeHandler.format.api]

    def get_format_data(
        self, message_type: str, data: dict
    ) -> Tuple[V20PresFormat, AttachDecorator]:
        """Get presentation format and attach objects for use in pres_ex messages."""

        return (
            V20PresFormat(
                attach_id=IndyPresExchangeHandler.format.api,
                format_=self.get_format_identifier(message_type),
            ),
            AttachDecorator.data_base64(data, ident=IndyPresExchangeHandler.format.api),
        )

    async def create_bound_request(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
    ) -> Tuple[V20PresFormat, AttachDecorator]:
        """Create a presentation request bound to a proposal.

        Args:
            pres_ex_record: Presentation exchange record for which
                to create presentation request
            request_data: Dict

        Returns:
            A tuple (updated presentation exchange record, presentation request message)

        """

        indy_proof_request = pres_ex_record.pres_proposal.attachment(
            IndyPresExchangeHandler.format
        )
        if request_data:
            indy_proof_request["name"] = request_data.get("name", "proof-request")
            indy_proof_request["version"] = request_data.get("version", "1.0")
            indy_proof_request["nonce"] = (
                request_data.get("nonce") or await generate_pr_nonce()
            )
        else:
            indy_proof_request["name"] = "proof-request"
            indy_proof_request["version"] = "1.0"
            indy_proof_request["nonce"] = await generate_pr_nonce()
        return self.get_format_data(PRES_20_REQUEST, indy_proof_request)

    async def create_pres(
        self,
        pres_ex_record: V20PresExRecord,
        request_data: Optional[dict] = None,
    ) -> Tuple[V20PresFormat, AttachDecorator]:
        """Create a presentation."""

        requested_credentials = {}
        if not request_data:
            try:
                proof_request = pres_ex_record.pres_request
                indy_proof_request = proof_request.attachment(
                    IndyPresExchangeHandler.format
                )
                requested_credentials = await indy_proof_req_preview2indy_requested_creds(
                    indy_proof_request,
                    preview=None,
                    holder=self._profile.inject(IndyHolder),
                )
            except ValueError as err:
                LOGGER.warning(f"{err}")
                raise V20PresFormatHandlerError(
                    f"No matching Indy credentials found: {err}"
                )
        else:
            if IndyPresExchangeHandler.format.api in request_data:
                indy_spec = request_data.get(IndyPresExchangeHandler.format.api)
                requested_credentials = {
                    "self_attested_attributes": indy_spec["self_attested_attributes"],
                    "requested_attributes": indy_spec["requested_attributes"],
                    "requested_predicates": indy_spec["requested_predicates"],
                }
        indy_handler = IndyPresExchHandler(self._profile)
        indy_proof = await indy_handler.return_presentation(
            pres_ex_record=pres_ex_record,
            requested_credentials=requested_credentials,
        )
        return self.get_format_data(PRES_20, indy_proof)

    async def receive_pres(self, message: V20Pres, pres_ex_record: V20PresExRecord):
        """Receive a presentation and check for presented values vs. proposal request."""

        def _check_proof_vs_proposal():
            """Check for bait and switch in presented values vs. proposal request."""
            proof_req = pres_ex_record.pres_request.attachment(
                IndyPresExchangeHandler.format
            )

            # revealed attrs
            for reft, attr_spec in proof["requested_proof"]["revealed_attrs"].items():
                proof_req_attr_spec = proof_req["requested_attributes"].get(reft)
                if not proof_req_attr_spec:
                    raise V20PresFormatHandlerError(
                        f"Presentation referent {reft} not in proposal request"
                    )
                req_restrictions = proof_req_attr_spec.get("restrictions", {})

                name = proof_req_attr_spec["name"]
                proof_value = attr_spec["raw"]
                sub_proof_index = attr_spec["sub_proof_index"]
                schema_id = proof["identifiers"][sub_proof_index]["schema_id"]
                cred_def_id = proof["identifiers"][sub_proof_index]["cred_def_id"]
                criteria = {
                    "schema_id": schema_id,
                    "schema_issuer_did": schema_id.split(":")[-4],
                    "schema_name": schema_id.split(":")[-2],
                    "schema_version": schema_id.split(":")[-1],
                    "cred_def_id": cred_def_id,
                    "issuer_did": cred_def_id.split(":")[-5],
                    f"attr::{name}::value": proof_value,
                }

                if (
                    not any(r.items() <= criteria.items() for r in req_restrictions)
                    and len(req_restrictions) != 0
                ):
                    raise V20PresFormatHandlerError(
                        f"Presented attribute {reft} does not satisfy proof request "
                        f"restrictions {req_restrictions}"
                    )

            # revealed attr groups
            for reft, attr_spec in (
                proof["requested_proof"].get("revealed_attr_groups", {}).items()
            ):
                proof_req_attr_spec = proof_req["requested_attributes"].get(reft)
                if not proof_req_attr_spec:
                    raise V20PresFormatHandlerError(
                        f"Presentation referent {reft} not in proposal request"
                    )
                req_restrictions = proof_req_attr_spec.get("restrictions", {})
                proof_values = {
                    name: values["raw"] for name, values in attr_spec["values"].items()
                }
                sub_proof_index = attr_spec["sub_proof_index"]
                schema_id = proof["identifiers"][sub_proof_index]["schema_id"]
                cred_def_id = proof["identifiers"][sub_proof_index]["cred_def_id"]
                criteria = {
                    "schema_id": schema_id,
                    "schema_issuer_did": schema_id.split(":")[-4],
                    "schema_name": schema_id.split(":")[-2],
                    "schema_version": schema_id.split(":")[-1],
                    "cred_def_id": cred_def_id,
                    "issuer_did": cred_def_id.split(":")[-5],
                    **{
                        f"attr::{name}::value": value
                        for name, value in proof_values.items()
                    },
                }

                if (
                    not any(r.items() <= criteria.items() for r in req_restrictions)
                    and len(req_restrictions) != 0
                ):
                    raise V20PresFormatHandlerError(
                        f"Presented attr group {reft} does not satisfy proof request "
                        f"restrictions {req_restrictions}"
                    )

            # predicate bounds
            for reft, pred_spec in proof["requested_proof"]["predicates"].items():
                proof_req_pred_spec = proof_req["requested_predicates"].get(reft)
                if not proof_req_pred_spec:
                    raise V20PresFormatHandlerError(
                        f"Presentation referent {reft} not in proposal request"
                    )
                req_name = proof_req_pred_spec["name"]
                req_pred = Predicate.get(proof_req_pred_spec["p_type"])
                req_value = proof_req_pred_spec["p_value"]
                req_restrictions = proof_req_pred_spec.get("restrictions", {})
                for req_restriction in req_restrictions:
                    for k in list(req_restriction):  # cannot modify en passant
                        if k.startswith("attr::"):
                            req_restriction.pop(k)  # let indy-sdk reject mismatch here
                sub_proof_index = pred_spec["sub_proof_index"]
                for ge_proof in proof["proof"]["proofs"][sub_proof_index][
                    "primary_proof"
                ]["ge_proofs"]:
                    proof_pred_spec = ge_proof["predicate"]
                    if proof_pred_spec["attr_name"] != canon(req_name):
                        continue
                    if not (
                        Predicate.get(proof_pred_spec["p_type"]) is req_pred
                        and proof_pred_spec["value"] == req_value
                    ):
                        raise V20PresFormatHandlerError(
                            f"Presentation predicate on {req_name} "
                            "mismatches proposal request"
                        )
                    break
                else:
                    raise V20PresFormatHandlerError(
                        f"Proposed request predicate on {req_name} not in presentation"
                    )

                schema_id = proof["identifiers"][sub_proof_index]["schema_id"]
                cred_def_id = proof["identifiers"][sub_proof_index]["cred_def_id"]
                criteria = {
                    "schema_id": schema_id,
                    "schema_issuer_did": schema_id.split(":")[-4],
                    "schema_name": schema_id.split(":")[-2],
                    "schema_version": schema_id.split(":")[-1],
                    "cred_def_id": cred_def_id,
                    "issuer_did": cred_def_id.split(":")[-5],
                }

                if (
                    not any(r.items() <= criteria.items() for r in req_restrictions)
                    and len(req_restrictions) != 0
                ):
                    raise V20PresFormatHandlerError(
                        f"Presented predicate {reft} does not satisfy proof request "
                        f"restrictions {req_restrictions}"
                    )

        proof = message.attachment(IndyPresExchangeHandler.format)
        _check_proof_vs_proposal()

    async def verify_pres(self, pres_ex_record: V20PresExRecord) -> V20PresExRecord:
        """Verify a presentation.

        Args:
            pres_ex_record: presentation exchange record
                with presentation request and presentation to verify

        Returns:
            presentation exchange record, updated

        """
        pres_request_msg = pres_ex_record.pres_request

        # The `or` anoncreds format is for the indy <--> anoncreds compatibility
        indy_proof_request = pres_request_msg.attachment(
            IndyPresExchangeHandler.format
        ) or pres_request_msg.attachment(AnonCredsPresExchangeHandler.format)
        indy_proof = pres_ex_record.pres.attachment(
            IndyPresExchangeHandler.format
        ) or pres_ex_record.pres.attachment(AnonCredsPresExchangeHandler.format)
        indy_handler = IndyPresExchHandler(self._profile)
        (
            schemas,
            cred_defs,
            rev_reg_defs,
            rev_reg_entries,
        ) = await indy_handler.process_pres_identifiers(indy_proof["identifiers"])

        verifier = self._profile.inject(IndyVerifier)
        (verified, verified_msgs) = await verifier.verify_presentation(
            indy_proof_request,
            indy_proof,
            schemas,
            cred_defs,
            rev_reg_defs,
            rev_reg_entries,
        )
        pres_ex_record.verified = json.dumps(verified)
        pres_ex_record.verified_msgs = list(set(verified_msgs))
        return pres_ex_record



"""Classes to manage credentials."""

import asyncio
import json
import logging
from typing import Mapping, Optional, Tuple

from ....cache.base import BaseCache
from ....connections.models.conn_record import ConnRecord
from ....core.error import BaseError
from ....core.profile import Profile
from ....indy.holder import IndyHolder, IndyHolderError
from ....indy.issuer import IndyIssuer, IndyIssuerRevocationRegistryFullError
from ....ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    GET_SCHEMA,
    IndyLedgerRequestsExecutor,
)
from ....messaging.credential_definitions.util import (
    CRED_DEF_SENT_RECORD_TYPE,
    CRED_DEF_TAGS,
)
from ....messaging.responder import BaseResponder
from ....multitenant.base import BaseMultitenantManager
from ....revocation.indy import IndyRevocation
from ....revocation.models.issuer_cred_rev_record import IssuerCredRevRecord
from ....revocation.models.revocation_registry import RevocationRegistry
from ....storage.base import BaseStorage
from ....storage.error import StorageError, StorageNotFoundError
from ...out_of_band.v1_0.models.oob_record import OobRecord
from .messages.credential_ack import CredentialAck
from .messages.credential_issue import CredentialIssue
from .messages.credential_offer import CredentialOffer
from .messages.credential_problem_report import (
    CredentialProblemReport,
    ProblemReportReason,
)
from .messages.credential_proposal import CredentialProposal
from .messages.credential_request import CredentialRequest
from .messages.inner.credential_preview import CredentialPreview
from .models.credential_exchange import V10CredentialExchange

LOGGER = logging.getLogger(__name__)


class CredentialManagerError(BaseError):
    """Credential error."""


class CredentialManager:
    """Class for managing credentials."""

    def __init__(self, profile: Profile):
        """Initialize a CredentialManager.

        Args:
            profile: The profile instance for this credential manager
        """
        self._profile = profile

    @property
    def profile(self) -> Profile:
        """Accessor for the current profile instance.

        Returns:
            The profile instance for this credential manager

        """
        return self._profile

    async def _match_sent_cred_def_id(self, tag_query: Mapping[str, str]) -> str:
        """Return most recent matching id of cred def that agent sent to ledger."""

        async with self._profile.session() as session:
            storage = session.inject(BaseStorage)
            found = await storage.find_all_records(
                type_filter=CRED_DEF_SENT_RECORD_TYPE, tag_query=tag_query
            )
        if not found:
            raise CredentialManagerError(
                f"Issuer has no operable cred def for proposal spec {tag_query}"
            )
        return max(found, key=lambda r: int(r.tags["epoch"])).tags["cred_def_id"]

    async def prepare_send(
        self,
        connection_id: str,
        credential_proposal: CredentialProposal,
        auto_remove: Optional[bool] = None,
        comment: Optional[str] = None,
    ) -> Tuple[V10CredentialExchange, CredentialOffer]:
        """Set up a new credential exchange for an automated send.

        Args:
            connection_id: Connection to create offer for
            credential_proposal: The credential proposal with preview
            auto_remove: Flag to automatically remove the record on completion
            comment: Optional human-readable comment to set in offer message

        Returns:
            A tuple of the new credential exchange record and credential offer message

        """
        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        credential_exchange = V10CredentialExchange(
            connection_id=connection_id,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            credential_proposal_dict=credential_proposal,
            auto_issue=True,
            auto_remove=auto_remove,
            trace=(credential_proposal._trace is not None),
        )
        (credential_exchange, credential_offer) = await self.create_offer(
            cred_ex_record=credential_exchange,
            counter_proposal=None,
            comment=comment,
        )
        return (credential_exchange, credential_offer)

    async def create_proposal(
        self,
        connection_id: str,
        *,
        auto_offer: Optional[bool] = None,
        auto_remove: Optional[bool] = None,
        comment: Optional[str] = None,
        credential_preview: Optional[CredentialPreview] = None,
        schema_id: Optional[str] = None,
        schema_issuer_did: Optional[str] = None,
        schema_name: Optional[str] = None,
        schema_version: Optional[str] = None,
        cred_def_id: Optional[str] = None,
        issuer_did: Optional[str] = None,
        trace: bool = False,
    ) -> V10CredentialExchange:
        """Create a credential proposal.

        Args:
            connection_id: Connection to create proposal for
            auto_offer: Should this proposal request automatically be handled to
                offer a credential
            auto_remove: Should the record be automatically removed on completion
            comment: Optional human-readable comment to include in proposal
            credential_preview: The credential preview to use to create
                the credential proposal
            schema_id: Schema id for credential proposal
            schema_issuer_did: Schema issuer DID for credential proposal
            schema_name: Schema name for credential proposal
            schema_version: Schema version for credential proposal
            cred_def_id: Credential definition id for credential proposal
            issuer_did: Issuer DID for credential proposal
            trace: Whether to trace the operation

        Returns:
            Resulting credential exchange record including credential proposal

        """
        credential_proposal_message = CredentialProposal(
            comment=comment,
            credential_proposal=credential_preview,
            schema_id=schema_id,
            schema_issuer_did=schema_issuer_did,
            schema_name=schema_name,
            schema_version=schema_version,
            cred_def_id=cred_def_id,
            issuer_did=issuer_did,
        )
        credential_proposal_message.assign_trace_decorator(self._profile.settings, trace)

        if auto_remove is None:
            auto_remove = not self._profile.settings.get("preserve_exchange_records")
        cred_ex_record = V10CredentialExchange(
            connection_id=connection_id,
            thread_id=credential_proposal_message._thread_id,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_PROPOSAL_SENT,
            credential_proposal_dict=credential_proposal_message,
            auto_offer=auto_offer,
            auto_remove=auto_remove,
            trace=trace,
        )
        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="create credential proposal")
        return cred_ex_record

    async def receive_proposal(
        self, message: CredentialProposal, connection_id: str
    ) -> V10CredentialExchange:
        """Receive a credential proposal.

        Returns:
            The resulting credential exchange record, created

        """
        # at this point, cred def and schema still open to potential negotiation
        cred_ex_record = V10CredentialExchange(
            connection_id=connection_id,
            thread_id=message._thread_id,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_PROPOSAL_RECEIVED,
            credential_proposal_dict=message,
            auto_offer=self._profile.settings.get(
                "debug.auto_respond_credential_proposal"
            ),
            auto_issue=self._profile.settings.get(
                "debug.auto_respond_credential_request"
            ),
            auto_remove=not self._profile.settings.get("preserve_exchange_records"),
            trace=(message._trace is not None),
        )
        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="receive credential proposal")

        return cred_ex_record

    async def create_offer(
        self,
        cred_ex_record: V10CredentialExchange,
        counter_proposal: Optional[CredentialProposal] = None,
        comment: Optional[str] = None,
    ) -> Tuple[V10CredentialExchange, CredentialOffer]:
        """Create a credential offer, update credential exchange record.

        Args:
            cred_ex_record: Credential exchange to create offer for
            counter_proposal: optional proposal to counter
            comment: optional human-readable comment to set in offer message

        Returns:
            A tuple (credential exchange record, credential offer message)

        """

        async def _create(cred_def_id):
            issuer = self._profile.inject(IndyIssuer)
            offer_json = await issuer.create_credential_offer(cred_def_id)
            return json.loads(offer_json)

        credential_proposal_message = (
            counter_proposal
            if counter_proposal
            else cred_ex_record.credential_proposal_dict
        )
        credential_proposal_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )
        cred_def_id = await self._match_sent_cred_def_id(
            {
                t: getattr(credential_proposal_message, t)
                for t in CRED_DEF_TAGS
                if getattr(credential_proposal_message, t)
            }
        )

        credential_preview = credential_proposal_message.credential_proposal

        # vet attributes
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]
        async with ledger:
            schema_id = await ledger.credential_definition_id2schema_id(cred_def_id)
            schema = await ledger.get_schema(schema_id)
        schema_attrs = set(schema["attrNames"])
        preview_attrs = set(credential_preview.attr_dict())
        if preview_attrs != schema_attrs:
            raise CredentialManagerError(
                f"Preview attributes {preview_attrs} "
                f"mismatch corresponding schema attributes {schema_attrs}"
            )

        credential_offer = None
        cache_key = f"credential_offer::{cred_def_id}"
        cache = self._profile.inject_or(BaseCache)
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    credential_offer = entry.result
                else:
                    credential_offer = await _create(cred_def_id)
                    await entry.set_result(credential_offer, 3600)
        if not credential_offer:
            credential_offer = await _create(cred_def_id)

        credential_offer_message = CredentialOffer(
            comment=comment,
            credential_preview=credential_preview,
            offers_attach=[CredentialOffer.wrap_indy_offer(credential_offer)],
        )

        credential_offer_message._thread = {"thid": credential_offer_message._thread_id}
        credential_offer_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        cred_ex_record.thread_id = credential_offer_message._thread_id
        cred_ex_record.schema_id = credential_offer["schema_id"]
        cred_ex_record.credential_definition_id = credential_offer["cred_def_id"]
        cred_ex_record.state = V10CredentialExchange.STATE_OFFER_SENT
        cred_ex_record.credential_proposal_dict = (  # any counter replaces original
            credential_proposal_message
        )
        cred_ex_record.credential_offer = credential_offer

        cred_ex_record.credential_offer_dict = credential_offer_message

        async with self._profile.session() as session:
            await cred_ex_record.save(session, reason="create credential offer")

        return (cred_ex_record, credential_offer_message)

    async def receive_offer(
        self, message: CredentialOffer, connection_id: Optional[str]
    ) -> V10CredentialExchange:
        """Receive a credential offer.

        Returns:
            The credential exchange record, updated

        """
        credential_preview = message.credential_preview
        indy_offer = message.indy_offer(0)
        schema_id = indy_offer["schema_id"]
        cred_def_id = indy_offer["cred_def_id"]

        credential_proposal_dict = CredentialProposal(
            comment=message.comment,
            credential_proposal=credential_preview,
            schema_id=schema_id,
            cred_def_id=cred_def_id,
        )

        async with self._profile.transaction() as txn:
            # Get credential exchange record (holder sent proposal first)
            # or create it (issuer sent offer first)
            try:
                cred_ex_record = await (
                    V10CredentialExchange.retrieve_by_connection_and_thread(
                        txn,
                        connection_id,
                        message._thread_id,
                        role=V10CredentialExchange.ROLE_HOLDER,
                        for_update=True,
                    )
                )
            except StorageNotFoundError:  # issuer sent this offer free of any proposal
                cred_ex_record = V10CredentialExchange(
                    connection_id=connection_id,
                    thread_id=message._thread_id,
                    initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
                    role=V10CredentialExchange.ROLE_HOLDER,
                    auto_remove=not self._profile.settings.get(
                        "preserve_exchange_records"
                    ),
                    trace=(message._trace is not None),
                )
            else:
                if cred_ex_record.state != V10CredentialExchange.STATE_PROPOSAL_SENT:
                    raise CredentialManagerError(
                        f"Credential exchange {cred_ex_record.credential_exchange_id} "
                        f"in {cred_ex_record.state} state "
                        f"(must be {V10CredentialExchange.STATE_PROPOSAL_SENT})"
                    )

            cred_ex_record.credential_proposal_dict = credential_proposal_dict
            cred_ex_record.credential_offer_dict = message
            cred_ex_record.credential_offer = indy_offer
            cred_ex_record.state = V10CredentialExchange.STATE_OFFER_RECEIVED
            cred_ex_record.schema_id = schema_id
            cred_ex_record.credential_definition_id = cred_def_id

            await cred_ex_record.save(txn, reason="receive credential offer")
            await txn.commit()

        return cred_ex_record

    async def create_request(
        self, cred_ex_record: V10CredentialExchange, holder_did: str
    ) -> Tuple[V10CredentialExchange, CredentialRequest]:
        """Create a credential request.

        Args:
            cred_ex_record: Credential exchange record
                for which to create request
            holder_did: holder DID

        Returns:
            A tuple (credential exchange record, credential request message)

        """
        credential_definition_id = cred_ex_record.credential_definition_id
        cred_offer_ser = cred_ex_record._credential_offer.ser
        cred_req_ser = None
        cred_req_meta = None

        # hold on to values that may have changed so we can restore after fetch
        auto_remove = cred_ex_record.auto_remove

        async def _create():
            multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
            else:
                ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
            ledger = (
                await ledger_exec_inst.get_ledger_for_identifier(
                    credential_definition_id,
                    txn_record_type=GET_CRED_DEF,
                )
            )[1]
            async with ledger:
                credential_definition = await ledger.get_credential_definition(
                    credential_definition_id
                )

            holder = self._profile.inject(IndyHolder)
            request_json, metadata_json = await holder.create_credential_request(
                cred_offer_ser,
                credential_definition,
                holder_did,
            )
            return {
                "request": json.loads(request_json),
                "metadata": json.loads(metadata_json),
            }

        if cred_ex_record.state == V10CredentialExchange.STATE_REQUEST_SENT:
            LOGGER.warning(
                "create_request called multiple times for v1.0 credential exchange: %s",
                cred_ex_record.credential_exchange_id,
            )
            cred_req_ser = cred_ex_record._credential_request.ser
            cred_req_meta = cred_ex_record.credential_request_metadata
        elif cred_ex_record.state == V10CredentialExchange.STATE_OFFER_RECEIVED:
            nonce = cred_offer_ser["nonce"]
            cache_key = (
                f"credential_request::{credential_definition_id}::{holder_did}::{nonce}"
            )
            cred_req_result = None
            cache = self._profile.inject_or(BaseCache)
            if cache:
                async with cache.acquire(cache_key) as entry:
                    if entry.result:
                        cred_req_result = entry.result
                    else:
                        cred_req_result = await _create()
                        await entry.set_result(cred_req_result, 3600)
            if not cred_req_result:
                cred_req_result = await _create()
            cred_req_ser = cred_req_result["request"]
            cred_req_meta = cred_req_result["metadata"]

            async with self._profile.transaction() as txn:
                cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                    txn, cred_ex_record.credential_exchange_id, for_update=True
                )
                if cred_ex_record.state != V10CredentialExchange.STATE_OFFER_RECEIVED:
                    raise CredentialManagerError(
                        f"Credential exchange {cred_ex_record.credential_exchange_id} "
                        f"in {cred_ex_record.state} state "
                        f"(must be {V10CredentialExchange.STATE_OFFER_RECEIVED})"
                    )

                cred_ex_record.credential_request = cred_req_ser
                cred_ex_record.credential_request_metadata = cred_req_meta
                cred_ex_record.state = V10CredentialExchange.STATE_REQUEST_SENT
                # restore values passed in...
                cred_ex_record.auto_remove = auto_remove
                await cred_ex_record.save(txn, reason="create credential request")
                await txn.commit()
        else:
            raise CredentialManagerError(
                f"Credential exchange {cred_ex_record.credential_exchange_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V10CredentialExchange.STATE_OFFER_RECEIVED})"
            )

        credential_request_message = CredentialRequest(
            requests_attach=[CredentialRequest.wrap_indy_cred_req(cred_req_ser)]
        )
        # Assign thid (and optionally pthid) to message
        credential_request_message.assign_thread_from(
            cred_ex_record.credential_offer_dict
        )
        credential_request_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        return (cred_ex_record, credential_request_message)

    async def receive_request(
        self,
        message: CredentialRequest,
        connection_record: Optional[ConnRecord],
        oob_record: Optional[OobRecord],
    ):
        """Receive a credential request.

        Args:
            message (CredentialRequest): The credential request message to receive.
            connection_record (Optional[ConnRecord]): The connection record associated
                with the request.
            oob_record (Optional[OobRecord]): The out-of-band record associated with the
                request.

        Returns:
            V10CredentialExchange: The credential exchange record, retrieved and updated.

        Raises:
            StorageNotFoundError: If the credential exchange record is not found.

        """
        assert len(message.requests_attach or []) == 1
        credential_request = message.indy_cred_req(0)

        # connection_id is None in the record if this is in response to
        # an request~attach from an OOB message. If so, we do not want to filter
        # the record by connection_id.
        connection_id = None if oob_record else connection_record.connection_id

        async with self._profile.transaction() as txn:
            try:
                cred_ex_record = await (
                    V10CredentialExchange.retrieve_by_connection_and_thread(
                        txn,
                        connection_id,
                        message._thread_id,
                        role=V10CredentialExchange.ROLE_ISSUER,
                        for_update=True,
                    )
                )
            except StorageNotFoundError as ex:
                LOGGER.error(
                    f"Credential Exchange (thread id = {message._thread_id}) not found."
                    " Indy issue credential format can't start from credential request.",
                )
                raise ex
            if cred_ex_record.state != V10CredentialExchange.STATE_OFFER_SENT:
                LOGGER.error(
                    "Skipping credential request; exchange state is %s (id=%s)",
                    cred_ex_record.state,
                    cred_ex_record.credential_exchange_id,
                )
                return None

            if connection_record:
                cred_ex_record.connection_id = connection_record.connection_id

            cred_ex_record.credential_request = credential_request
            cred_ex_record.state = V10CredentialExchange.STATE_REQUEST_RECEIVED
            await cred_ex_record.save(txn, reason="receive credential request")
            await txn.commit()

        return cred_ex_record

    async def issue_credential(
        self,
        cred_ex_record: V10CredentialExchange,
        *,
        comment: Optional[str] = None,
        retries: int = 5,
    ) -> Tuple[V10CredentialExchange, CredentialIssue]:
        """Issue a credential.

        Args:
            cred_ex_record: The credential exchange record
                for which to issue a credential
            comment: optional human-readable comment pertaining to credential issue
            retries: how many times to retry on error

        Returns:
            Tuple: (Updated credential exchange record, credential message)

        """

        credential_ser = None

        if cred_ex_record.credential:
            LOGGER.warning(
                "issue_credential called multiple times for v1.0 credential exchange %s",
                cred_ex_record.credential_exchange_id,
            )
            credential_ser = cred_ex_record._credential.ser

        elif cred_ex_record.state != V10CredentialExchange.STATE_REQUEST_RECEIVED:
            raise CredentialManagerError(
                f"Credential exchange {cred_ex_record.credential_exchange_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V10CredentialExchange.STATE_REQUEST_RECEIVED})"
            )

        else:
            cred_offer_ser = cred_ex_record._credential_offer.ser
            cred_req_ser = cred_ex_record._credential_request.ser
            cred_values = (
                cred_ex_record.credential_proposal_dict.credential_proposal.attr_dict(
                    decode=False
                )
            )
            schema_id = cred_ex_record.schema_id
            cred_def_id = cred_ex_record.credential_definition_id

            issuer = self.profile.inject(IndyIssuer)
            multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
            else:
                ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
            ledger = (
                await ledger_exec_inst.get_ledger_for_identifier(
                    schema_id,
                    txn_record_type=GET_SCHEMA,
                )
            )[1]
            async with ledger:
                schema = await ledger.get_schema(schema_id)
                credential_definition = await ledger.get_credential_definition(
                    cred_ex_record.credential_definition_id
                )
            revocable = credential_definition["value"].get("revocation")

            for attempt in range(max(retries, 1)):
                if attempt > 0:
                    LOGGER.info(
                        "Waiting 2s before retrying credential issuance "
                        "for cred def '%s'",
                        cred_def_id,
                    )
                    await asyncio.sleep(2)

                if revocable:
                    revoc = IndyRevocation(self._profile)
                    registry_info = await revoc.get_or_create_active_registry(cred_def_id)
                    if not registry_info:
                        continue
                    del revoc
                    issuer_rev_reg, rev_reg = registry_info
                    rev_reg_id = issuer_rev_reg.revoc_reg_id
                    tails_path = rev_reg.tails_local_path
                else:
                    rev_reg_id = None
                    tails_path = None

                try:
                    (credential_json, cred_rev_id) = await issuer.create_credential(
                        schema,
                        cred_offer_ser,
                        cred_req_ser,
                        cred_values,
                        rev_reg_id,
                        tails_path,
                    )
                except IndyIssuerRevocationRegistryFullError:
                    # unlucky, another instance filled the registry first
                    continue

                if revocable and rev_reg.max_creds <= int(cred_rev_id):
                    revoc = IndyRevocation(self._profile)
                    await revoc.handle_full_registry(rev_reg_id)
                    del revoc

                credential_ser = json.loads(credential_json)
                break

            if not credential_ser:
                raise CredentialManagerError(
                    f"Cred def id {cred_ex_record.credential_definition_id} "
                    "has no active revocation registry"
                ) from None

            async with self._profile.transaction() as txn:
                if revocable and cred_rev_id:
                    issuer_cr_rec = IssuerCredRevRecord(
                        state=IssuerCredRevRecord.STATE_ISSUED,
                        cred_ex_id=cred_ex_record.credential_exchange_id,
                        cred_ex_version=IssuerCredRevRecord.VERSION_1,
                        rev_reg_id=rev_reg_id,
                        cred_rev_id=cred_rev_id,
                    )
                    await issuer_cr_rec.save(
                        txn,
                        reason=(
                            "Created issuer cred rev record for "
                            f"rev reg id {rev_reg_id}, index {cred_rev_id}"
                        ),
                    )

                cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                    txn, cred_ex_record.credential_exchange_id, for_update=True
                )
                if cred_ex_record.state != V10CredentialExchange.STATE_REQUEST_RECEIVED:
                    raise CredentialManagerError(
                        f"Credential exchange {cred_ex_record.credential_exchange_id} "
                        f"in {cred_ex_record.state} state "
                        f"(must be {V10CredentialExchange.STATE_REQUEST_RECEIVED})"
                    )
                cred_ex_record.state = V10CredentialExchange.STATE_ISSUED
                cred_ex_record.credential = credential_ser
                cred_ex_record.revoc_reg_id = rev_reg_id
                cred_ex_record.revocation_id = cred_rev_id
                await cred_ex_record.save(txn, reason="issue credential")
                await txn.commit()

        credential_message = CredentialIssue(
            comment=comment,
            credentials_attach=[CredentialIssue.wrap_indy_credential(credential_ser)],
        )
        credential_message._thread = {"thid": cred_ex_record.thread_id}
        credential_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        return (cred_ex_record, credential_message)

    async def receive_credential(
        self, message: CredentialIssue, connection_id: Optional[str]
    ) -> V10CredentialExchange:
        """Receive a credential from an issuer.

        Hold in storage potentially to be processed by controller before storing.

        Returns:
            Credential exchange record, retrieved and updated

        """
        assert len(message.credentials_attach or []) == 1
        raw_credential = message.indy_credential(0)

        async with self._profile.transaction() as txn:
            try:
                cred_ex_record = await (
                    V10CredentialExchange.retrieve_by_connection_and_thread(
                        txn,
                        connection_id,
                        message._thread_id,
                        role=V10CredentialExchange.ROLE_HOLDER,
                        for_update=True,
                    )
                )
            except StorageNotFoundError:
                raise CredentialManagerError(
                    "No credential exchange record found for received credential"
                ) from None
            if cred_ex_record.state != V10CredentialExchange.STATE_REQUEST_SENT:
                raise CredentialManagerError(
                    f"Credential exchange {cred_ex_record.credential_exchange_id} "
                    f"in {cred_ex_record.state} state "
                    f"(must be {V10CredentialExchange.STATE_REQUEST_SENT})"
                )
            cred_ex_record.raw_credential = raw_credential
            cred_ex_record.state = V10CredentialExchange.STATE_CREDENTIAL_RECEIVED

            await cred_ex_record.save(txn, reason="receive credential")
            await txn.commit()

        return cred_ex_record

    async def store_credential(
        self, cred_ex_record: V10CredentialExchange, credential_id: Optional[str] = None
    ) -> V10CredentialExchange:
        """Store a credential in holder wallet; send ack to issuer.

        Args:
            cred_ex_record: credential exchange record
                with credential to store and ack
            credential_id: optional credential identifier to override default on storage

        Returns:
            Updated credential exchange record

        """
        if cred_ex_record.state != V10CredentialExchange.STATE_CREDENTIAL_RECEIVED:
            raise CredentialManagerError(
                f"Credential exchange {cred_ex_record.credential_exchange_id} "
                f"in {cred_ex_record.state} state "
                f"(must be {V10CredentialExchange.STATE_CREDENTIAL_RECEIVED})"
            )

        raw_cred_serde = cred_ex_record._raw_credential
        revoc_reg_def = None
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                raw_cred_serde.de.cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]
        async with ledger:
            credential_definition = await ledger.get_credential_definition(
                raw_cred_serde.de.cred_def_id
            )
            if raw_cred_serde.de.rev_reg_id:
                revoc_reg_def = await ledger.get_revoc_reg_def(
                    raw_cred_serde.de.rev_reg_id
                )

        holder = self._profile.inject(IndyHolder)
        if (
            cred_ex_record.credential_proposal_dict
            and cred_ex_record.credential_proposal_dict.credential_proposal
        ):
            mime_types = (
                cred_ex_record.credential_proposal_dict.credential_proposal.mime_types()
            )
        else:
            mime_types = None

        if revoc_reg_def:
            revoc_reg = RevocationRegistry.from_definition(revoc_reg_def, True)
            await revoc_reg.get_or_fetch_local_tails_path()
        try:
            credential_id = await holder.store_credential(
                credential_definition,
                raw_cred_serde.ser,
                cred_ex_record.credential_request_metadata,
                mime_types,
                credential_id=credential_id,
                rev_reg_def=revoc_reg_def,
            )
        except IndyHolderError as e:
            LOGGER.error("Error storing credential: %s: %s", e.error_code, e.message)
            raise e

        credential_json = await holder.get_credential(credential_id)
        credential = json.loads(credential_json)

        async with self._profile.transaction() as txn:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                txn, cred_ex_record.credential_exchange_id, for_update=True
            )
            if cred_ex_record.state != V10CredentialExchange.STATE_CREDENTIAL_RECEIVED:
                raise CredentialManagerError(
                    f"Credential exchange {cred_ex_record.credential_exchange_id} "
                    f"in {cred_ex_record.state} state "
                    f"(must be {V10CredentialExchange.STATE_CREDENTIAL_RECEIVED})"
                )

            cred_ex_record.credential_id = credential_id
            cred_ex_record.credential = credential
            cred_ex_record.revoc_reg_id = credential.get("rev_reg_id", None)
            cred_ex_record.revocation_id = credential.get("cred_rev_id", None)
            await cred_ex_record.save(txn, reason="store credential")
            await txn.commit()

        return cred_ex_record

    async def send_credential_ack(
        self,
        cred_ex_record: V10CredentialExchange,
    ) -> Tuple[V10CredentialExchange, CredentialAck]:
        """Create, send, and return ack message for input credential exchange record.

        Delete credential exchange record if set to auto-remove.

        Returns:
            a tuple of the updated credential exchange record
            and the credential ack message for tracing

        """
        credential_ack_message = CredentialAck()
        credential_ack_message.assign_thread_id(
            cred_ex_record.thread_id, cred_ex_record.parent_thread_id
        )
        credential_ack_message.assign_trace_decorator(
            self._profile.settings, cred_ex_record.trace
        )

        try:
            async with self._profile.transaction() as txn:
                try:
                    cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                        txn, cred_ex_record.credential_exchange_id, for_update=True
                    )
                except StorageNotFoundError:
                    LOGGER.warning(
                        "Skipping credential exchange ack, record not found: '%s'",
                        cred_ex_record.credential_exchange_id,
                    )
                    return (cred_ex_record, None)

                if (
                    cred_ex_record.state
                    != V10CredentialExchange.STATE_CREDENTIAL_RECEIVED
                ):
                    LOGGER.warning(
                        "Skipping credential exchange ack, state is '%s' for record '%s'",
                        cred_ex_record.state,
                        cred_ex_record.credential_exchange_id,
                    )
                    return (cred_ex_record, None)

                cred_ex_record.state = V10CredentialExchange.STATE_ACKED
                await cred_ex_record.save(txn, reason="ack credential")
                await txn.commit()

            if cred_ex_record.auto_remove:
                async with self._profile.session() as session:
                    await cred_ex_record.delete_record(session)  # all done: delete

        except StorageError:
            LOGGER.exception(
                "Error updating credential exchange"
            )  # holder still owes an ack: carry on

        responder = self._profile.inject_or(BaseResponder)
        if responder:
            await responder.send_reply(
                credential_ack_message,
                connection_id=cred_ex_record.connection_id,
            )
        else:
            LOGGER.warning(
                "Configuration has no BaseResponder: cannot ack credential on %s",
                cred_ex_record.thread_id,
            )

        return (cred_ex_record, credential_ack_message)

    async def receive_credential_ack(
        self, message: CredentialAck, connection_id: Optional[str]
    ) -> Optional[V10CredentialExchange]:
        """Receive credential ack from holder.

        Returns:
            credential exchange record, retrieved and updated

        """
        async with self._profile.transaction() as txn:
            try:
                cred_ex_record = await (
                    V10CredentialExchange.retrieve_by_connection_and_thread(
                        txn,
                        connection_id,
                        message._thread_id,
                        role=V10CredentialExchange.ROLE_ISSUER,
                        for_update=True,
                    )
                )
            except StorageNotFoundError:
                LOGGER.warning(
                    "Skip ack message on credential exchange, record not found %s",
                    message._thread_id,
                )
                return None

            if cred_ex_record.state == V10CredentialExchange.STATE_ACKED:
                return None
            cred_ex_record.state = V10CredentialExchange.STATE_ACKED
            await cred_ex_record.save(txn, reason="credential acked")
            await txn.commit()

        if cred_ex_record.auto_remove:
            async with self._profile.session() as session:
                await cred_ex_record.delete_record(session)  # all done: delete

        return cred_ex_record

    async def receive_problem_report(
        self, message: CredentialProblemReport, connection_id: str
    ):
        """Receive problem report.

        Returns:
            credential exchange record, retrieved and updated

        """
        async with self._profile.transaction() as txn:
            try:
                cred_ex_record = await (
                    V10CredentialExchange.retrieve_by_connection_and_thread(
                        txn, connection_id, message._thread_id, for_update=True
                    )
                )
            except StorageNotFoundError:
                LOGGER.warning(
                    "Skip problem report on credential exchange, record not found %s",
                    message._thread_id,
                )
                return None

            cred_ex_record.state = V10CredentialExchange.STATE_ABANDONED
            code = message.description.get(
                "code",
                ProblemReportReason.ISSUANCE_ABANDONED.value,
            )
            cred_ex_record.error_msg = f"{code}: {message.description.get('en', code)}"
            await cred_ex_record.save(txn, reason="received problem report")
            await txn.commit()

        return cred_ex_record



from .base import (
    DEFAULT_LOGGING_CONFIG_PATH_INI,
    DEFAULT_MULTITENANT_LOGGING_CONFIG_PATH_INI,
    LOG_FORMAT_FILE_ALIAS_PATTERN,
)
from .configurator import LoggingConfigurator, fileConfig, load_resource
from .filters import ContextFilter, context_wallet_id
from .timed_rotating_file_multi_process_handler import (
    TimedRotatingFileMultiProcessHandler,
)

__all__ = [
    "DEFAULT_LOGGING_CONFIG_PATH_INI",
    "DEFAULT_MULTITENANT_LOGGING_CONFIG_PATH_INI",
    "LOG_FORMAT_FILE_ALIAS_PATTERN",
    "LoggingConfigurator",
    "load_resource",
    "fileConfig",
    "ContextFilter",
    "context_wallet_id",
    "TimedRotatingFileMultiProcessHandler",
]



"""LinkedDataProof."""

from typing import Optional

from marshmallow import INCLUDE, fields, post_dump

from ....messaging.models.base import BaseModel, BaseModelSchema
from ....messaging.valid import (
    ISO8601_DATETIME_EXAMPLE,
    ISO8601_DATETIME_VALIDATE,
    UUID4_EXAMPLE,
    Uri,
)


class LDProof(BaseModel):
    """Linked Data Proof model."""

    class Meta:
        """LinkedDataProof metadata."""

        schema_class = "LinkedDataProofSchema"

    def __init__(
        self,
        type: Optional[str] = None,
        proof_purpose: Optional[str] = None,
        verification_method: Optional[str] = None,
        created: Optional[str] = None,
        domain: Optional[str] = None,
        challenge: Optional[str] = None,
        jws: Optional[str] = None,
        proof_value: Optional[str] = None,
        nonce: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Initialize the LDProof instance."""

        self.type = type
        self.proof_purpose = proof_purpose
        self.verification_method = verification_method
        self.created = created
        self.domain = domain
        self.challenge = challenge
        self.jws = jws
        self.proof_value = proof_value
        self.nonce = nonce
        self.extra = kwargs


class LinkedDataProofSchema(BaseModelSchema):
    """Linked data proof schema.

    Based on https://w3c-ccg.github.io/ld-proofs

    """

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = LDProof

    type = fields.Str(
        required=True,
        metadata={
            "description": (
                "Identifies the digital signature suite that was used to create the"
                " signature"
            ),
            "example": "Ed25519Signature2018",
        },
    )

    proof_purpose = fields.Str(
        data_key="proofPurpose",
        required=True,
        metadata={"description": "Proof purpose", "example": "assertionMethod"},
    )

    verification_method = fields.Str(
        data_key="verificationMethod",
        required=True,
        validate=Uri(),
        metadata={
            "description": "Information used for proof verification",
            "example": (
                "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg34"
                "2Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
            ),
        },
    )

    created = fields.Str(
        required=False,
        validate=ISO8601_DATETIME_VALIDATE,
        metadata={
            "description": (
                "The string value of an ISO8601 combined date and time string generated"
                " by the Signature Algorithm"
            ),
            "example": ISO8601_DATETIME_EXAMPLE,
        },
    )

    domain = fields.Str(
        required=False,
        metadata={
            "description": (
                "A string value specifying the restricted domain of the signature."
            ),
            "example": "https://example.com",
        },
    )

    challenge = fields.Str(
        required=False,
        metadata={
            "description": (
                "Associates a challenge with a proof, for use with a proofPurpose such"
                " as authentication"
            ),
            "example": UUID4_EXAMPLE,
        },
    )

    jws = fields.Str(
        required=False,
        metadata={
            "description": "Associates a Detached Json Web Signature with a proof",
            "example": (
                "eyJhbGciOiAiRWREUc2UsICJjcml0IjogWyJiNjQiXX0..lKJU0Df_keblRKhZAS9"
                "Qq6zybm-HqUXNVZ8vgEPNTAjQ1Ch6YBKY7UBAjg6iBX5qBQ"
            ),
        },
    )

    proof_value = fields.Str(
        required=False,
        data_key="proofValue",
        metadata={
            "description": "The proof value of a proof",
            "example": (
                "sy1AahqbzJQ63n9RtekmwzqZeVj494VppdAVJBnMYrTwft6cLJJGeTSSxCCJ6HKnR"
                "twE7jjDh6sB2z2AAiZY9BBnCD8wUVgwqH3qchGRCuC2RugA4eQ9fUrR4Yuycac3caiaaay"
            ),
        },
    )

    nonce = fields.Str(
        required=False,
        metadata={
            "description": "The nonce",
            "example": (
                "CF69iO3nfvqRsRBNElE8b4wO39SyJHPM7Gg1nExltW5vSfQA1lvDCR/zXX1To0/4NLo=="
            ),
        },
    )

    @post_dump(pass_original=True)
    def add_unknown_properties(self, data: dict, original, **kwargs):
        """Add back unknown properties before outputting."""

        data.update(original.extra)

        return data



"""Profile manager for multitenancy."""

import logging

from ..config.base import InjectionError
from ..config.injector import BaseInjector
from ..config.provider import BaseProvider
from ..config.settings import BaseSettings
from ..utils.classloader import ClassLoader, ClassNotFoundError

LOGGER = logging.getLogger(__name__)


class MultitenantManagerProvider(BaseProvider):
    """Multitenant manager provider.

    Decides which manager to use based on the settings.
    """

    single_wallet_askar_manager_path = (
        "acapy_agent.multitenant."
        "single_wallet_askar_manager.SingleWalletAskarMultitenantManager"
    )
    MANAGER_TYPES = {
        "basic": "acapy_agent.multitenant.manager.MultitenantManager",
        "single-wallet-askar": single_wallet_askar_manager_path,
    }

    def __init__(self, root_profile):
        """Initialize the multitenant profile manager provider."""
        self.root_profile = root_profile
        self._inst = {}

    def provide(self, settings: BaseSettings, injector: BaseInjector):
        """Create the multitenant manager instance."""

        manager_type = settings.get_value(
            "multitenant.wallet_type", default="basic"
        ).lower()

        manager_class = self.MANAGER_TYPES.get(manager_type, manager_type)

        if manager_class not in self._inst:
            LOGGER.info("Create multitenant manager: %s", manager_type)
            try:
                self._inst[manager_class] = ClassLoader.load_class(manager_class)(
                    self.root_profile
                )
            except ClassNotFoundError as err:
                raise InjectionError(
                    f"Unknown multitenant manager type: {manager_type}"
                ) from err

        return self._inst[manager_class]



"""Presentation message handler."""

from .....core.oob_processor import OobMessageProcessor
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import PresentationManager
from ..messages.presentation import Presentation
from ..messages.presentation_problem_report import ProblemReportReason


class PresentationHandler(BaseHandler):
    """Message handler class for presentations."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for presentations.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()
        profile = context.profile
        self._logger.debug("PresentationHandler called with context %s", context)
        assert isinstance(context.message, Presentation)
        self._logger.debug(
            "Received presentation message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for presentation not ready")

        # Find associated oob record. If the presentation request was created as an oob
        # attachment the presentation exchange record won't have a connection id (yet)
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Normally we would do a check here that there is either a connection or
        # an associated oob record. However as present proof supported receiving
        # presentation without oob record or connection record
        # (aip-1 style connectionless) we can't perform this check here

        presentation_manager = PresentationManager(profile)

        presentation_exchange_record = await presentation_manager.receive_presentation(
            context.message, context.connection_record, oob_record
        )  # mgr saves record state null if need be and possible

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="PresentationHandler.handle.END",
            perf_counter=r_time,
        )

        # Automatically move to next state if flag is set
        if (
            presentation_exchange_record
            and presentation_exchange_record.auto_verify
            or context.settings.get("debug.auto_verify_presentation")
        ):
            try:
                await presentation_manager.verify_presentation(
                    presentation_exchange_record, responder
                )
            except (BaseModelError, LedgerError, StorageError) as err:
                self._logger.exception(err)
                if presentation_exchange_record:
                    async with profile.session() as session:
                        await presentation_exchange_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            presentation_exchange_record,
                            ProblemReportReason.ABANDONED.value,  # them: be vague
                        )
                    )

            trace_event(
                context.settings,
                presentation_exchange_record,
                outcome="PresentationHandler.handle.VERIFY",
                perf_counter=r_time,
            )



import asyncio
from unittest import IsolatedAsyncioTestCase

from ..message import InboundMessage
from ..receipt import MessageReceipt


class TestInboundMessage(IsolatedAsyncioTestCase):
    async def test_wait_response(self):
        message = InboundMessage(
            payload="test",
            connection_id="conn_id",
            receipt=MessageReceipt(),
            session_id="session_id",
        )
        assert not message.processing_complete_event.is_set()
        message.dispatch_processing_complete()
        assert message.processing_complete_event.is_set()

        message = InboundMessage(
            payload="test",
            connection_id="conn_id",
            receipt=MessageReceipt(),
            session_id="session_id",
        )
        assert not message.processing_complete_event.is_set()
        task = message.wait_processing_complete()
        message.dispatch_processing_complete()
        await asyncio.wait_for(task, 1)



"""Wallet base class."""

from abc import ABC, abstractmethod
from typing import List, Optional, Sequence, Tuple, Union

from ..ledger.base import BaseLedger
from ..ledger.endpoint_type import EndpointType
from .did_info import DIDInfo, KeyInfo
from .did_method import SOV, DIDMethod
from .error import WalletError
from .key_type import KeyType


class BaseWallet(ABC):
    """Abstract wallet interface."""

    @abstractmethod
    async def create_signing_key(
        self,
        key_type: KeyType,
        seed: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> KeyInfo:
        """Create a new public/private signing keypair.

        Args:
            key_type: Key type to create
            seed: Optional seed allowing deterministic key creation
            metadata: Optional metadata to store with the keypair

        Returns:
            A `KeyInfo` representing the new record

        """

    @abstractmethod
    async def create_key(
        self,
        key_type: KeyType,
        seed: Optional[str] = None,
        metadata: Optional[dict] = None,
        kid: Optional[str] = None,
    ) -> KeyInfo:
        """Create a new public/private keypair.

        Args:
            key_type: Key type to create
            seed: Seed for key
            metadata: Optional metadata to store with the keypair
            kid: Optional key identifier

        Returns:
            A `KeyInfo` representing the new record

        Raises:
            WalletDuplicateError: If the resulting verkey already exists in the wallet
            WalletError: If there is another backend error
        """

    @abstractmethod
    async def assign_kid_to_key(self, verkey: str, kid: str) -> KeyInfo:
        """Assign a KID to a key.

        This is separate from the create_key method because some DIDs are only
        known after keys are created.

        Args:
            verkey: The verification key of the keypair
            kid: The kid to assign to the keypair

        Returns:
            A `KeyInfo` representing the keypair

        """

    @abstractmethod
    async def get_key_by_kid(self, kid: str) -> KeyInfo:
        """Fetch a key by looking up its kid.

        Args:
            kid: the key identifier

        Returns:
            The key identified by kid

        """

    @abstractmethod
    async def get_signing_key(self, verkey: str) -> KeyInfo:
        """Fetch info for a signing keypair.

        Args:
            verkey: The verification key of the keypair

        Returns:
            A `KeyInfo` representing the keypair

        """

    @abstractmethod
    async def replace_signing_key_metadata(self, verkey: str, metadata: dict):
        """Replace the metadata associated with a signing keypair.

        Args:
            verkey: The verification key of the keypair
            metadata: The new metadata to store

        """

    @abstractmethod
    async def rotate_did_keypair_start(
        self, did: str, next_seed: Optional[str] = None
    ) -> str:
        """Begin key rotation for DID that wallet owns: generate new keypair.

        Args:
            did: signing DID
            next_seed: seed for incoming ed25519 key pair (default random)

        Returns:
            The new verification key

        Raises:
            WalletNotFoundError: if wallet does not own DID

        """

    @abstractmethod
    async def rotate_did_keypair_apply(self, did: str) -> None:
        """Apply temporary keypair as main for DID that wallet owns.

        Args:
            did: signing DID

        Raises:
            WalletNotFoundError: if wallet does not own DID
            WalletError: if wallet has not started key rotation

        """

    @abstractmethod
    async def create_local_did(
        self,
        method: DIDMethod,
        key_type: KeyType,
        seed: Optional[str] = None,
        did: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> DIDInfo:
        """Create and store a new local DID.

        Args:
            method: The method to use for the DID
            key_type: The key type to use for the DID
            seed: Optional seed to use for DID
            did: The DID to use
            metadata: Metadata to store with DID

        Returns:
            The created `DIDInfo`

        """

    @abstractmethod
    async def store_did(self, did_info: DIDInfo) -> DIDInfo:
        """Store a DID in the wallet.

        This enables components external to the wallet to define how a DID
        is created and then store it in the wallet for later use.

        Args:
            did_info: The DID to store

        Returns:
            The stored `DIDInfo`
        """

    async def create_public_did(
        self,
        method: DIDMethod,
        key_type: KeyType,
        seed: Optional[str] = None,
        did: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> DIDInfo:
        """Create and store a new public DID.

        This method creates a new public DID using the specified DID method and key type.

        The optional `seed` parameter can be used to provide a seed for the DID
            generation.

        If a `did` is provided, it will be used as the DID instead of generating a new
            one.

        The `metadata` parameter can be used to store additional metadata with the DID.

        Args:
            method: The DID method to use for creating the DID.
            key_type: The key type to use for the DID.
            seed: Optional seed to use for DID generation.
            did: The DID to use instead of generating a new one.
            metadata: Optional metadata to store with the DID.

        Returns:
            The created `DIDInfo` object.

        """
        metadata = metadata or {}
        metadata.setdefault("posted", True)
        did_info = await self.create_local_did(
            method=method, key_type=key_type, seed=seed, did=did, metadata=metadata
        )
        return await self.set_public_did(did_info)

    @abstractmethod
    async def get_public_did(self) -> DIDInfo | None:
        """Retrieve the public DID.

        Returns:
            The currently public `DIDInfo`, if any

        """

    @abstractmethod
    async def set_public_did(self, did: Union[str, DIDInfo]) -> DIDInfo:
        """Assign the public DID.

        Returns:
            The updated `DIDInfo`

        """

    @abstractmethod
    async def get_local_dids(self) -> Sequence[DIDInfo]:
        """Get list of defined local DIDs.

        Returns:
            A list of `DIDInfo` instances

        """

    @abstractmethod
    async def get_local_did(self, did: str) -> DIDInfo:
        """Find info for a local DID.

        Args:
            did: The DID for which to get info

        Returns:
            A `DIDInfo` instance for the DID

        """

    @abstractmethod
    async def get_local_did_for_verkey(self, verkey: str) -> DIDInfo:
        """Resolve a local DID from a verkey.

        Args:
            verkey: Verkey for which to get DID info

        Returns:
            A `DIDInfo` instance for the DID

        """

    @abstractmethod
    async def replace_local_did_metadata(self, did: str, metadata: dict):
        """Replace the metadata associated with a local DID.

        Prefer `set_did_endpoint()` to set endpoint in metadata.

        Args:
            did: DID for which to replace metadata
            metadata: The new metadata

        """

    async def get_posted_dids(self) -> Sequence[DIDInfo]:
        """Get list of defined posted DIDs.

        Returns:
            A list of `DIDInfo` instances

        """
        return [
            info for info in await self.get_local_dids() if info.metadata.get("posted")
        ]

    async def set_did_endpoint(
        self,
        did: str,
        endpoint: str,
        _ledger: BaseLedger,
        endpoint_type: Optional[EndpointType] = None,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
        routing_keys: Optional[List[str]] = None,
    ):
        """Update the endpoint for a DID in the wallet, send to ledger if posted.

        Args:
            did (str): The DID for which to set the endpoint.
            endpoint (str): The endpoint to set. Use None to clear the endpoint.
            _ledger (BaseLedger): The ledger to which to send the endpoint update if the
                DID is public or posted.
            endpoint_type (EndpointType, optional): The type of the endpoint/service.
                Only endpoint_type 'endpoint' affects the local wallet.
            write_ledger (bool, optional): Whether to write the endpoint update to the
                ledger. Defaults to True.
            endorser_did (str, optional): The DID of the endorser. Defaults to None.
            routing_keys (List[str], optional): The list of routing keys.
                Defaults to None.

        Raises:
            WalletError: If the DID method is not 'did:sov'.

        """
        did_info = await self.get_local_did(did)

        if did_info.method != SOV:
            raise WalletError("Setting DID endpoint is only allowed for did:sov DIDs")
        metadata = {**did_info.metadata}
        if not endpoint_type:
            endpoint_type = EndpointType.ENDPOINT
        if endpoint_type == EndpointType.ENDPOINT:
            metadata[endpoint_type.indy] = endpoint

        await self.replace_local_did_metadata(did, metadata)

    @abstractmethod
    async def sign_message(
        self, message: Union[List[bytes], bytes], from_verkey: str
    ) -> bytes:
        """Sign message(s) using the private key associated with a given verkey.

        Args:
            message: The message(s) to sign
            from_verkey: Sign using the private key related to this verkey

        Returns:
            The signature

        """

    @abstractmethod
    async def verify_message(
        self,
        message: Union[List[bytes], bytes],
        signature: bytes,
        from_verkey: str,
        key_type: KeyType,
    ) -> bool:
        """Verify a signature against the public key of the signer.

        Args:
            message: The message to verify
            signature: The signature to verify
            from_verkey: Verkey to use in verification
            key_type: The key type to derive the signature verification algorithm from

        Returns:
            True if verified, else False

        """

    @abstractmethod
    async def pack_message(
        self, message: str, to_verkeys: Sequence[str], from_verkey: Optional[str] = None
    ) -> bytes:
        """Pack a message for one or more recipients.

        Args:
            message: The message to pack
            to_verkeys: The verkeys to pack the message for
            from_verkey: The sender verkey

        Returns:
            The packed message

        """

    @abstractmethod
    async def unpack_message(self, enc_message: bytes) -> Tuple[str, str, str]:
        """Unpack a message.

        Args:
            enc_message: The encrypted message

        Returns:
            A tuple: (message, from_verkey, to_verkey)

        """

    def __repr__(self) -> str:
        """Get a human readable string."""
        return "<{}>".format(self.__class__.__name__)



"""Handler for incoming query messages."""

from .....messaging.base_handler import (
    BaseHandler,
    BaseResponder,
    HandlerException,
    RequestContext,
)
from ..manager import V10DiscoveryMgr
from ..messages.query import Query


class QueryHandler(BaseHandler):
    """Handler for incoming query messages."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler implementation."""
        self._logger.debug("QueryHandler called with context %s", context)
        assert isinstance(context.message, Query)

        if not context.connection_ready:
            raise HandlerException("No connection established")

        profile = context.profile
        mgr = V10DiscoveryMgr(profile)
        reply = await mgr.receive_query(context.message)
        reply.assign_thread_from(context.message)
        reply.assign_trace_from(context.message)
        await responder.send_reply(reply)



from unittest import TestCase

from ..predicate import Predicate


class TestPredicate(TestCase):
    """Predicate tests for coverage"""

    def test_get_monikers(self):
        """Get predicate."""
        assert Predicate.get("LT") == Predicate.get("$lt") == Predicate.get("<")
        assert Predicate.get("LE") == Predicate.get("$lte") == Predicate.get("<=")
        assert Predicate.get("GE") == Predicate.get("$gte") == Predicate.get(">=")
        assert Predicate.get("GT") == Predicate.get("$gt") == Predicate.get(">")
        assert Predicate.get("!=") is None

        assert Predicate.get("LT").fortran == "LT"
        assert Predicate.get("LT").wql == "$lt"
        assert Predicate.get("LT").math == "<"

    def test_cmp(self):
        """Test comparison via predicates"""
        assert Predicate.get("LT").value.yes(0, 1)
        assert Predicate.get("LT").value.yes("0", "1")
        assert Predicate.get("LT").value.no(0, 0)
        assert Predicate.get("LT").value.no(1, 0)
        assert Predicate.get("LT").value.no("1", "0")
        assert Predicate.get("LT").value.no("0", "0")

        assert Predicate.get("LE").value.yes(0, 1)
        assert Predicate.get("LE").value.yes("0", "1")
        assert Predicate.get("LE").value.yes(0, 0)
        assert Predicate.get("LE").value.no(1, 0)
        assert Predicate.get("LE").value.no("1", "0")
        assert Predicate.get("LE").value.yes("0", "0")

        assert Predicate.get("GE").value.no(0, 1)
        assert Predicate.get("GE").value.no("0", "1")
        assert Predicate.get("GE").value.yes(0, 0)
        assert Predicate.get("GE").value.yes(1, 0)
        assert Predicate.get("GE").value.yes("1", "0")
        assert Predicate.get("GE").value.yes("0", "0")

        assert Predicate.get("GT").value.no(0, 1)
        assert Predicate.get("GT").value.no("0", "1")
        assert Predicate.get("GT").value.no(0, 0)
        assert Predicate.get("GT").value.yes(1, 0)
        assert Predicate.get("GT").value.yes("1", "0")
        assert Predicate.get("GT").value.no("0", "0")



import asyncio
from unittest import IsolatedAsyncioTestCase

from ...connections.models.conn_record import ConnRecord
from ...storage.base import BaseStorage
from ...storage.record import StorageRecord
from ...tests import mock
from ...utils.testing import create_test_profile
from ...version import __version__
from ...wallet.models.wallet_record import WalletRecord
from .. import upgrade as test_module
from ..upgrade import UpgradeError


class TestUpgrade(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            record = StorageRecord(
                "acapy_version",
                "v0.7.2",
            )
            await storage.add_record(record)
        recs = [
            WalletRecord(
                key_management_mode=[
                    WalletRecord.MODE_UNMANAGED,
                    WalletRecord.MODE_MANAGED,
                ][i],
                settings={
                    "wallet.name": f"my-wallet-{i}",
                    "wallet.type": "indy",
                    "wallet.key": f"dummy-wallet-key-{i}",
                },
                wallet_name=f"my-wallet-{i}",
            )
            for i in range(2)
        ]
        async with self.profile.session() as session:
            for rec in recs:
                await rec.save(session)

    def test_bad_calls(self):
        with self.assertRaises(SystemExit):
            test_module.execute(["bad"])

    async def test_upgrade_storage_from_version_included(self):
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
        ):
            await test_module.upgrade(
                settings={
                    "upgrade.config_path": "./acapy_agent/commands/default_version_upgrade_config.yml",
                    "upgrade.from_version": "v0.7.2",
                }
            )

    async def test_upgrade_storage_missing_from_version(self):
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
        ):
            await test_module.upgrade(settings={})

    async def test_upgrade_from_version(self):
        self.profile.settings.extend(
            {
                "upgrade.from_version": "v0.7.2",
            }
        )
        with (
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
        ):
            await test_module.upgrade(
                profile=self.profile,
            )

    async def test_upgrade_all_subwallets(self):
        self.profile.settings.extend(
            {
                "upgrade.from_version": "v0.7.2",
                "upgrade.upgrade_all_subwallets": True,
                "upgrade.force_upgrade": True,
                "upgrade.page_size": 1,
            }
        )
        with (
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
        ):
            await test_module.upgrade(
                profile=self.profile,
            )

    async def test_upgrade_specified_subwallets(self):
        wallet_ids = []
        async with self.profile.session() as session:
            wallet_recs = await WalletRecord.query(session, tag_filter={})
        for wallet_rec in wallet_recs:
            wallet_ids.append(wallet_rec.wallet_id)
        self.profile.settings.extend(
            {
                "upgrade.named_tags": "fix_issue_rev_reg",
                "upgrade.upgrade_subwallets": [wallet_ids[0]],
                "upgrade.force_upgrade": True,
            }
        )
        with (
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
        ):
            await test_module.upgrade(
                profile=self.profile,
            )

        self.profile.settings.extend(
            {
                "upgrade.named_tags": "fix_issue_rev_reg",
                "upgrade.upgrade_subwallets": wallet_ids,
                "upgrade.force_upgrade": True,
                "upgrade.page_size": 1,
            }
        )
        with (
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
        ):
            await test_module.upgrade(
                profile=self.profile,
            )

    async def test_upgrade_callable(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            version_storage_record = await storage.find_record(
                type_filter="acapy_version", tag_query={}
            )
            await storage.delete_record(version_storage_record)
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                test_module.yaml,
                "safe_load",
                mock.MagicMock(
                    return_value={
                        "v0.7.2": {
                            "resave_records": {
                                "base_record_path": [
                                    "acapy_agent.connections.models.conn_record.ConnRecord"
                                ]
                            },
                            "update_existing_records": True,
                        },
                    }
                ),
            ),
        ):
            await test_module.upgrade(
                settings={
                    "upgrade.from_version": "v0.7.2",
                }
            )

    async def test_upgrade_callable_named_tag(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            version_storage_record = await storage.find_record(
                type_filter="acapy_version", tag_query={}
            )
            await storage.delete_record(version_storage_record)
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                test_module.yaml,
                "safe_load",
                mock.MagicMock(
                    return_value={
                        "v0.7.2": {
                            "resave_records": {
                                "base_record_path": [
                                    "acapy_agent.connections.models.conn_record.ConnRecord"
                                ]
                            },
                            "update_existing_records": True,
                        },
                        "fix_issue_rev_reg": {"fix_issue_rev_reg_records": True},
                    }
                ),
            ),
        ):
            await test_module.upgrade(
                settings={
                    "upgrade.named_tags": ["fix_issue_rev_reg"],
                    "upgrade.force_upgrade": True,
                }
            )

    async def test_upgrade_x_same_version(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            version_storage_record = await storage.find_record(
                type_filter="acapy_version", tag_query={}
            )
            await storage.update_record(version_storage_record, f"v{__version__}", {})
        with mock.patch.object(
            test_module,
            "wallet_config",
            mock.CoroutineMock(
                return_value=(
                    self.profile,
                    mock.CoroutineMock(did="public DID", verkey="verkey"),
                )
            ),
        ):
            await test_module.upgrade(
                settings={
                    "upgrade.config_path": "./acapy_agent/commands/default_version_upgrade_config.yml",
                }
            )

    async def test_upgrade_missing_from_version(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            version_storage_record = await storage.find_record(
                type_filter="acapy_version", tag_query={}
            )
            await storage.delete_record(version_storage_record)
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
        ):
            with self.assertRaises(UpgradeError) as ctx:
                await test_module.upgrade(
                    settings={
                        "upgrade.config_path": "./acapy_agent/commands/default_version_upgrade_config.yml",
                    }
                )
            assert "Error during upgrade: No upgrade from version or tags found" in str(
                ctx.exception
            )

    async def test_upgrade_x_callable_not_set(self):
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            version_storage_record = await storage.find_record(
                type_filter="acapy_version", tag_query={}
            )
            await storage.delete_record(version_storage_record)
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                test_module.yaml,
                "safe_load",
                mock.MagicMock(
                    return_value={
                        "v0.7.2": {
                            "resave_records": {
                                "base_record_path": [
                                    "acapy_agent.connections.models.conn_record.ConnRecord"
                                ]
                            },
                            "update_existing_records": True,
                        },
                        "v0.6.0": {"update_existing_records_b": True},
                    }
                ),
            ),
        ):
            with self.assertRaises(UpgradeError) as ctx:
                await test_module.upgrade(
                    settings={
                        "upgrade.from_version": "v0.6.0",
                    }
                )
            assert "No function specified for" in str(ctx.exception)

    async def test_upgrade_x_class_not_found(self):
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                test_module.yaml,
                "safe_load",
                mock.MagicMock(
                    return_value={
                        "v0.7.2": {
                            "resave_records": {
                                "base_record_path": [
                                    "acapy_agent.connections.models.conn_record.Invalid"
                                ],
                            }
                        },
                    }
                ),
            ),
        ):
            with self.assertRaises(UpgradeError) as ctx:
                await test_module.upgrade(
                    settings={
                        "upgrade.from_version": "v0.7.2",
                    }
                )
            assert "Unknown Record type" in str(ctx.exception)

    async def test_execute(self):
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                ConnRecord,
                "query",
                mock.CoroutineMock(return_value=[ConnRecord()]),
            ),
            mock.patch.object(ConnRecord, "save", mock.CoroutineMock()),
            mock.patch.object(
                asyncio, "get_event_loop", mock.MagicMock()
            ) as mock_get_event_loop,
            mock.patch.object(
                # Normally, this would be a CoroutingMock. However, the coroutine
                # is awaited by run_until_complete, which is mocked out.
                # Use MagicMock to prevent unawaited coroutine warnings.
                test_module,
                "upgrade",
                mock.MagicMock(),
            ),
        ):
            mock_get_event_loop.return_value = mock.MagicMock(
                run_until_complete=mock.MagicMock(),
            )
            test_module.execute(
                [
                    "--upgrade-config",
                    "./acapy_agent/config/tests/test-acapy-upgrade-config.yaml",
                    "--from-version",
                    "v0.7.0",
                    "--force-upgrade",
                    "--wallet-test",
                ]
            )

    async def test_upgrade_x_invalid_record_type(self):
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                test_module.yaml,
                "safe_load",
                mock.MagicMock(
                    return_value={
                        "v0.7.2": {
                            "resave_records": {
                                "base_exch_record_path": [
                                    "acapy_agent.connections.models.connection_target.ConnectionTarget"
                                ],
                            }
                        }
                    }
                ),
            ),
        ):
            with self.assertRaises(UpgradeError) as ctx:
                await test_module.upgrade(
                    settings={
                        "upgrade.from_version": "v0.7.2",
                    }
                )
            assert "Only BaseRecord can be resaved" in str(ctx.exception)

    async def test_upgrade_force(self):
        with (
            mock.patch.object(
                test_module,
                "wallet_config",
                mock.CoroutineMock(
                    return_value=(
                        self.profile,
                        mock.CoroutineMock(did="public DID", verkey="verkey"),
                    )
                ),
            ),
            mock.patch.object(
                test_module.yaml,
                "safe_load",
                mock.MagicMock(
                    return_value={
                        "v0.7.2": {
                            "resave_records": {
                                "base_record_path": [
                                    "acapy_agent.connections.models.conn_record.ConnRecord"
                                ],
                            },
                            "update_existing_records": True,
                        },
                        "v0.7.3": {
                            "update_existing_records": True,
                        },
                        "v0.7.1": {
                            "update_existing_records": False,
                        },
                    }
                ),
            ),
        ):
            await test_module.upgrade(
                settings={
                    "upgrade.from_version": "v0.7.0",
                    "upgrade.force_upgrade": True,
                }
            )

    async def test_get_upgrade_version_list(self):
        assert len(test_module.get_upgrade_version_list(from_version="v0.7.2")) >= 1

    async def test_add_version_record(self):
        await test_module.add_version_record(self.profile, "v0.7.4")
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            version_storage_record = await storage.find_record(
                type_filter="acapy_version", tag_query={}
            )
            assert version_storage_record.value == "v0.7.4"
            await storage.delete_record(version_storage_record)
            with self.assertRaises(test_module.StorageNotFoundError):
                await storage.find_record(type_filter="acapy_version", tag_query={})
            await test_module.add_version_record(self.profile, "v0.7.5")
            version_storage_record = await storage.find_record(
                type_filter="acapy_version", tag_query={}
            )
            assert version_storage_record.value == "v0.7.5"

    async def test_upgrade_x_invalid_config(self):
        with mock.patch.object(
            test_module.yaml,
            "safe_load",
            mock.MagicMock(return_value={}),
        ):
            with self.assertRaises(UpgradeError) as ctx:
                await test_module.upgrade(profile=self.profile)
            assert "No version configs found in" in str(ctx.exception)

    async def test_upgrade_x_params(self):
        with self.assertRaises(UpgradeError) as ctx:
            await test_module.upgrade(profile=self.profile, settings={})
        assert "upgrade requires either profile or settings" in str(ctx.exception)
        with self.assertRaises(UpgradeError) as ctx:
            await test_module.upgrade(profile=self.profile, settings={"...": "..."})
        assert "upgrade requires either profile or settings" in str(ctx.exception)

    def test_main(self):
        with (
            mock.patch.object(test_module, "__name__", "__main__"),
            mock.patch.object(test_module, "execute", mock.MagicMock()) as mock_execute,
        ):
            test_module.main()
            mock_execute.assert_called_once

    def test_get_explicit_upgrade_option(self):
        assert not test_module.ExplicitUpgradeOption.get("test")
        assert (
            test_module.ExplicitUpgradeOption.get("critical")
            == test_module.ExplicitUpgradeOption.ERROR_AND_STOP
        )
        assert (
            test_module.ExplicitUpgradeOption.get("warning")
            == test_module.ExplicitUpgradeOption.LOG_AND_PROCEED
        )

    async def test_upgrade_explicit_upgrade_required_check(self):
        test_config_dict = {
            "v0.8.1": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
            },
            "v0.7.2": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
            },
            "v0.7.1": {"resave_records": [], "update_existing_records": False},
            "v0.7.0": {"resave_records": [], "update_existing_records": False},
            "v0.6.0": {"resave_records": [], "update_existing_records": False},
        }
        check, to_skip_list, exp_version = test_module.explicit_upgrade_required_check(
            ["v0.6.0", "v0.7.0", "v0.7.1", "v0.7.2", "v0.8.1"],
            test_config_dict,
        )
        assert (check, to_skip_list, exp_version) == (False, [], None)

        check, to_skip_list, exp_version = test_module.explicit_upgrade_required_check(
            ["v0.8.1", "v0.6.0", "v0.7.1", "v0.7.2", "v0.7.0"],
            test_config_dict,
        )
        assert (check, to_skip_list, exp_version) == (False, [], None)

        test_config_dict = {
            "v0.8.1": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
            },
            "v0.7.2": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
                "explicit_upgrade": "warning",
            },
            "v0.7.1": {
                "resave_records": [],
                "update_existing_records": False,
                "explicit_upgrade": "warning",
            },
            "v0.7.0": {"resave_records": [], "update_existing_records": False},
            "v0.6.0": {
                "resave_records": [],
                "update_existing_records": False,
                "explicit_upgrade": "warning",
            },
        }
        check, to_skip_list, exp_version = test_module.explicit_upgrade_required_check(
            ["v0.6.0", "v0.7.0", "v0.7.1", "v0.7.2", "v0.8.1"],
            test_config_dict,
        )
        assert (check, to_skip_list, exp_version) == (
            False,
            ["v0.6.0", "v0.7.1", "v0.7.2"],
            None,
        )

        test_config_dict = {
            "v0.8.1": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
            },
            "v0.7.2": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
                "explicit_upgrade": "critical",
            },
            "v0.7.1": {"resave_records": [], "update_existing_records": False},
            "v0.7.0": {"resave_records": [], "update_existing_records": False},
            "v0.6.0": {"resave_records": [], "update_existing_records": False},
        }
        check, to_skip_list, exp_version = test_module.explicit_upgrade_required_check(
            ["v0.6.0", "v0.7.0", "v0.7.1", "v0.7.2", "v0.8.1"],
            test_config_dict,
        )
        assert (check, to_skip_list, exp_version) == (True, [], "v0.7.2")

        test_config_dict = {
            "v0.8.1": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
            },
            "v0.7.2": {
                "resave_records": [
                    "acapy_agent.connections.models.conn_record.ConnRecord"
                ],
                "update_existing_records": False,
                "explicit_upgrade": "warning",
            },
            "v0.7.1": {
                "resave_records": [],
                "update_existing_records": False,
                "explicit_upgrade": "warning",
            },
            "v0.7.0": {
                "resave_records": [],
                "update_existing_records": False,
                "explicit_upgrade": "critical",
            },
            "v0.6.0": {"resave_records": [], "update_existing_records": False},
        }
        check, to_skip_list, exp_version = test_module.explicit_upgrade_required_check(
            ["v0.6.0", "v0.7.0", "v0.7.1", "v0.7.2", "v0.8.1"],
            test_config_dict,
        )
        assert (check, to_skip_list, exp_version) == (True, [], "v0.7.0")

    async def test_upgrade_explicit_check(self):
        self.profile.settings.extend(
            {
                "upgrade.from_version": "v0.7.0",
            }
        )
        with mock.patch.object(
            test_module.yaml,
            "safe_load",
            mock.MagicMock(
                return_value={
                    "v0.7.2": {
                        "resave_records": {
                            "base_record_path": [
                                "acapy_agent.connections.models.conn_record.ConnRecord"
                            ],
                        },
                        "update_existing_records": True,
                    },
                    "v0.7.3": {
                        "update_existing_records": True,
                        "explicit_upgrade": "critical",
                    },
                    "v0.7.1": {
                        "update_existing_records": False,
                    },
                }
            ),
        ):
            with self.assertRaises(UpgradeError) as ctx:
                await test_module.upgrade(profile=self.profile)
            assert "Explicit upgrade flag with critical value found" in str(ctx.exception)

        with mock.patch.object(
            test_module.yaml,
            "safe_load",
            mock.MagicMock(
                return_value={
                    "v0.7.2": {
                        "resave_records": {
                            "base_record_path": [
                                "acapy_agent.connections.models.conn_record.ConnRecord"
                            ],
                        },
                        "update_existing_records": True,
                        "explicit_upgrade": "warning",
                    },
                    "v0.7.3": {
                        "update_existing_records": True,
                        "explicit_upgrade": "critical",
                    },
                    "v0.7.1": {
                        "update_existing_records": False,
                    },
                }
            ),
        ):
            with self.assertRaises(UpgradeError) as ctx:
                await test_module.upgrade(profile=self.profile)
            assert "Explicit upgrade flag with critical value found" in str(ctx.exception)

        with (
            mock.patch.object(test_module, "LOGGER", mock.MagicMock()) as mock_logger,
            mock.patch.object(
                test_module.yaml,
                "safe_load",
                mock.MagicMock(
                    return_value={
                        "v0.7.2": {
                            "resave_records": {
                                "base_record_path": [
                                    "acapy_agent.connections.models.conn_record.ConnRecord"
                                ],
                            },
                            "update_existing_records": True,
                        },
                        "v0.7.3": {
                            "update_existing_records": True,
                            "explicit_upgrade": "warning",
                        },
                        "v0.7.1": {
                            "update_existing_records": True,
                            "explicit_upgrade": "warning",
                        },
                    }
                ),
            ),
        ):
            await test_module.upgrade(profile=self.profile)
            assert mock_logger.warning.call_count == 1
            assert mock_logger.info.call_count == 0



"""Controller proof purpose class."""

from typing import TYPE_CHECKING

from pyld import jsonld
from pyld.jsonld import JsonLdProcessor

from ..constants import SECURITY_CONTEXT_URL
from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from ..validation_result import PurposeResult
from .proof_purpose import ProofPurpose

# Avoid circular dependency
if TYPE_CHECKING:
    from ..suites import LinkedDataProof


class ControllerProofPurpose(ProofPurpose):
    """Controller proof purpose class."""

    def validate(
        self,
        *,
        proof: dict,
        document: dict,
        suite: "LinkedDataProof",
        verification_method: dict,
        document_loader: DocumentLoaderMethod,
    ) -> PurposeResult:
        """Validate whether verification method of proof is authorized by controller."""
        try:
            result = super().validate(
                proof=proof,
                document=document,
                suite=suite,
                verification_method=verification_method,
                document_loader=document_loader,
            )

            # Return early if super check was invalid
            if not result.valid:
                return result

            verification_id = verification_method.get("id")
            controller = verification_method.get("controller")

            if isinstance(controller, dict):
                controller_id = controller.get("id")
            elif isinstance(controller, str):
                controller_id = controller
            else:
                raise LinkedDataProofException('"controller" must be a string or dict')

            # Get the controller
            result.controller = jsonld.frame(
                controller_id,
                frame={
                    "@context": SECURITY_CONTEXT_URL,
                    "id": controller_id,
                    self.term: {"@embed": "@never", "id": verification_id},
                },
                options={
                    "documentLoader": document_loader,
                    "expandContext": SECURITY_CONTEXT_URL,
                    # if we don't set base explicitly it will remove the base in returned
                    # document (e.g. use key:z... instead of did:key:z...)
                    # same as compactToRelative in jsonld.js
                    "base": None,
                },
            )

            # Retrieve al verification methods on controller associated with term
            verification_methods = JsonLdProcessor.get_values(
                result.controller, self.term
            )

            # Check if any of the verification methods matches with the verification id
            result.valid = any(
                method == verification_id for method in verification_methods
            )

            if not result.valid:
                raise LinkedDataProofException(
                    f"Verification method {verification_id} not authorized"
                    f" by controller for proof purpose {self.term}"
                )

            return result

        except Exception as e:
            return PurposeResult(valid=False, error=str(e))



"""Classes to manage discover features."""

import asyncio
import logging
from typing import Optional

from ....core.error import BaseError
from ....core.profile import Profile
from ....core.protocol_registry import ProtocolRegistry
from ....messaging.responder import BaseResponder
from ....storage.error import StorageNotFoundError
from .messages.disclose import Disclose
from .messages.query import Query
from .models.discovery_record import V10DiscoveryExchangeRecord as DiscRecord


class V10DiscoveryMgrError(BaseError):
    """Discover feature v1_0 error."""


class V10DiscoveryMgr:
    """Class for discover feature v1_0 under RFC 31."""

    def __init__(self, profile: Profile):
        """Initialize a V10DiscoveryMgr.

        Args:
            profile: The profile for this manager
        """
        self._profile = profile
        self._logger = logging.getLogger(__name__)

    @property
    def profile(self) -> Profile:
        """Accessor for the current Profile.

        Returns:
            The Profile for this manager

        """
        return self._profile

    async def receive_disclose(
        self, disclose_msg: Disclose, connection_id: str
    ) -> DiscRecord:
        """Receive Disclose message and return updated V10DiscoveryExchangeRecord."""
        async with self._profile.session() as session:
            record = None
            if disclose_msg._thread:
                thread_id = disclose_msg._thread.thid
                try:
                    record = await DiscRecord.retrieve_by_id(
                        session=session, record_id=thread_id
                    )
                except StorageNotFoundError:
                    pass

            if not record:
                record = await DiscRecord.retrieve_if_exists_by_connection_id(
                    session, connection_id
                )

            if not record:
                record = DiscRecord()

            record.connection_id = connection_id
            record.disclose = disclose_msg
            record.state = DiscRecord.STATE_DISCLOSE_RECV
            await record.save(session)

        return record

    async def lookup_exchange_rec_by_connection(
        self, connection_id: str
    ) -> Optional[DiscRecord]:
        """Retrieve V20DiscoveryExchangeRecord by connection_id."""
        async with self._profile.session() as session:
            if await DiscRecord.exists_for_connection_id(
                session=session, connection_id=connection_id
            ):
                return await DiscRecord.retrieve_by_connection_id(
                    session=session, connection_id=connection_id
                )
            else:
                return None

    async def receive_query(self, query_msg: Query) -> Disclose:
        """Process query and return the corresponding disclose message."""
        registry = self._profile.context.inject(ProtocolRegistry)
        query_str = query_msg.query
        published_results = []
        protocols = registry.protocols_matching_query(query_str)
        results = await registry.prepare_disclosed(self._profile.context, protocols)

        async with self._profile.session() as session:
            to_publish_protocols = None
            if (
                session.settings.get("disclose_protocol_list")
                and len(session.settings.get("disclose_protocol_list")) > 0
            ):
                to_publish_protocols = session.settings.get("disclose_protocol_list")

            for result in results:
                to_publish_result = {}
                if "pid" in result:
                    if (
                        to_publish_protocols
                        and result.get("pid") not in to_publish_protocols
                    ):
                        continue

                    to_publish_result["pid"] = result.get("pid")
                else:
                    continue

                if "roles" in result:
                    to_publish_result["roles"] = result.get("roles")

                published_results.append(to_publish_result)

        disclose_msg = Disclose(protocols=published_results)
        # Check if query message has a thid
        # If disclosing this agents feature
        if query_msg._thread:
            disclose_msg.assign_thread_id(query_msg._thread.thid)
        return disclose_msg

    async def check_if_disclosure_received(self, record_id: str) -> DiscRecord:
        """Check if disclosures has been received."""
        while True:
            async with self._profile.session() as session:
                ex_rec = await DiscRecord.retrieve_by_id(
                    session=session, record_id=record_id
                )
            if ex_rec.disclose:
                return ex_rec
            await asyncio.sleep(0.5)

    async def create_and_send_query(
        self,
        query: str,
        comment: Optional[str] = None,
        connection_id: Optional[str] = None,
    ) -> DiscRecord:
        """Create and send a Query message."""
        query_msg = Query(query=query, comment=comment)
        if not connection_id:
            # Disclose this agent's features and/or goal codes
            record = DiscRecord()
            record.query_msg = query_msg
            disclose = await self.receive_query(query_msg=query_msg)
            record.disclose = disclose
            return record

        record = None
        async with self._profile.session() as session:
            record = await DiscRecord.retrieve_if_exists_by_connection_id(
                session=session, connection_id=connection_id
            )

            if record:
                record.disclose = None
            else:
                record = DiscRecord()

        record.state = DiscRecord.STATE_QUERY_SENT
        record.query_msg = query_msg
        record.connection_id = connection_id

        query_msg.assign_thread_id(record.discovery_exchange_id)
        responder = self._profile.inject(BaseResponder)

        await responder.send(query_msg, connection_id=connection_id)

        async with self._profile.session() as session:
            await record.save(session)

        try:
            return await asyncio.wait_for(
                self.check_if_disclosure_received(
                    record_id=record.discovery_exchange_id,
                ),
                5,
            )

        except asyncio.TimeoutError:
            return record
